
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Tags;                          use Ada.Tags;

package body Associations is

    function Create_Association return A_Association is
        this : constant A_Association := new Association;
    begin
        this.Construct;
        return this;
    end Create_Association;

    ----------------------------------------------------------------------------

    overriding
    procedure Adjust( this : access Association ) is
        crsr : Value_Maps.Cursor;
        v    : A_Value;
    begin
        Object(this.all).Adjust;
        this.pairs := Value_Maps.Map'(this.pairs);
        crsr := this.pairs.First;
        while Value_Maps.Has_Element( crsr ) loop
            v := Copy( Value_Maps.Element( crsr ) );
            this.pairs.Replace_Element( crsr, v );
            crsr := Value_Maps.Next( crsr );
        end loop;
    end Adjust;

    ----------------------------------------------------------------------------

    not overriding
    function As_Boolean( this : access Association; name : String ) return Boolean is
        crsr : constant Value_Maps.Cursor := this.pairs.Find( name );
    begin
        if Value_Maps.Has_Element( crsr ) then
           return Element( crsr ).As_Boolean;
        end if;
        raise VALUE_NOT_FOUND with "Value name '" & name & "' not in association";
    end As_Boolean;

    ----------------------------------------------------------------------------

    not overriding
    function As_Integer( this : access Association; name : String ) return Integer is
        crsr : constant Value_Maps.Cursor := this.pairs.Find( name );
    begin
        if Value_Maps.Has_Element( crsr ) then
           return Element( crsr ).As_Integer;
        end if;
        raise VALUE_NOT_FOUND with "Value name '" & name & "' not in association";
    end As_Integer;

    ----------------------------------------------------------------------------

    not overriding
    function As_String( this : access Association; name : String ) return String is
        crsr : constant Value_Maps.Cursor := this.pairs.Find( name );
    begin
        if Has_Element( crsr ) then
            return Value_Maps.Element( crsr ).As_String;
        end if;
        raise VALUE_NOT_FOUND with "Value name '" & name & "' not in association";
    end As_String;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Association ) is
        crsr : Value_Maps.Cursor;
        v    : A_Value;
    begin
        while not this.pairs.Is_Empty loop
            crsr := this.pairs.First;
            v := Element( crsr );
            Delete( v );
            this.pairs.Delete( crsr );
        end loop;
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    function Is_Defined( this : access Association; name : String ) return Boolean is
    begin
        return Has_Element( this.pairs.Find( name ) );
    end Is_Defined;

    ----------------------------------------------------------------------------

    not overriding
    procedure Iterate( this    : access Association;
                       examine : not null access procedure( name  : String;
                                                            value : A_Value ) ) is

        procedure Local_Examine( pos : Value_Maps.Cursor ) is
        begin
            examine.all( Key( pos ), Element( pos ) );
        end Local_Examine;

    begin
        this.pairs.Iterate( Local_Examine'Access );
    end Iterate;

    ----------------------------------------------------------------------------

    not overriding
    function Length( this : access Association ) return Natural is
    begin
        return Natural(this.pairs.Length);
    end Length;

    ----------------------------------------------------------------------------

    not overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Association is
        this : aliased Association;
    begin
        this.Construct;
        Association'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class;
                           obj    : out Association ) is

        procedure Read_Pair is
            name : constant String := String'Input( stream );
            val  : A_Value := A_Value'Input( stream );
        begin
            obj.Set_Value( name, val );
        end Read_Pair;

        len : Integer;
    begin
        Object'Read( stream, Object(obj) );
        len := Integer'Input( stream );
        for i in 1..len loop
            Read_Pair;
        end loop;
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class;
                            obj    : Association ) is

        procedure Write_Pair( pos : Value_Maps.Cursor ) is
        begin
            String'Output( stream, Key( pos ) );
            A_Value'Output( stream, Element( pos ) );
        end Write_Pair;

    begin
        Object'Write( stream, Object(obj) );
        Integer'Output( stream, Integer(obj.pairs.Length) );
        obj.pairs.Iterate( Write_Pair'Access );
    end Object_Write;

    ----------------------------------------------------------------------------

    not overriding
    procedure Remove( this : access Association; name : String ) is
        crsr : Value_Maps.Cursor := this.pairs.Find( name );
        val  : A_Value;
    begin
        if Has_Element( crsr ) then
            val := Element( crsr );
            Delete( val );
            this.pairs.Delete( crsr );
        end if;
    end Remove;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Default( this : access Association; name : String; val : Boolean ) is
        v : A_Value;
    begin
        if not this.Is_Defined( name ) then
            v := Create_Value( val );
            this.Set_Value( name, v );
        end if;
    end Set_Default;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Default( this : access Association; name : String; val : Integer ) is
        v : A_Value;
    begin
        if not this.Is_Defined( name ) then
            v := Create_Value( val );
            this.Set_Value( name, v );
        end if;
    end Set_Default;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Default( this : access Association; name : String; val : String ) is
        v : A_Value;
    begin
        if not this.Is_Defined( name ) then
            v := Create_Value( val );
            this.Set_Value( name, v );
        end if;
    end Set_Default;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Default( this : access Association;
                           name : String;
                           val  : in out A_Value ) is
    begin
        if not this.Is_Defined( name ) then
            this.Set_Value( name, val );
        else
            Delete( val );
        end if;
    end Set_Default;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Value( this : access Association; name : String; val : Boolean ) is
        v : A_Value := Create_Value( val );
    begin
        this.Set_Value( name, v );
    end Set_Value;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Value( this : access Association; name : String; val : Integer ) is
        v : A_Value := Create_Value( val );
    begin
        this.Set_Value( name, v );
    end Set_Value;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Value( this : access Association; name : String; val : String ) is
        v : A_Value := Create_Value( val );
    begin
        this.Set_Value( name, v );
    end Set_Value;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Value( this : access Association;
                         name : String;
                         val  : in out A_Value ) is
        crsr : constant Value_Maps.Cursor := this.pairs.Find( name );
        old  : A_Value;
    begin
        if Has_Element( crsr ) then
            old := Element( crsr );
            Delete( old );
            this.pairs.Replace_Element( crsr, val );
        else
            this.pairs.Insert( name, val );
        end if;
        val := null;
    end Set_Value;

    ----------------------------------------------------------------------------

    overriding
    function To_String( this : access Association ) return String is
        str  : Unbounded_String := To_Unbounded_String( "{" );
        crsr : Value_Maps.Cursor := this.pairs.First;
    begin
        while Has_Element( crsr ) loop
            str := str & Key( crsr ) & ":'" & Element( crsr ).As_String & "'";
            if Has_Element( Next( crsr ) ) then
                str := str & ",";
            end if;
            crsr := Next( crsr );
        end loop;
        return To_String( str ) & "}";
    end To_String;

    ----------------------------------------------------------------------------

    function A_Association_Input( stream : access Root_Stream_Type'Class ) return A_Association is
        obj : A_Association := null;
    begin
        if Boolean'Input( stream ) then
            A_Association'Read( stream, obj );
        end if;
        return obj;
    end A_Association_Input;

    ----------------------------------------------------------------------------

    procedure A_Association_Output( stream : access Root_Stream_Type'Class;
                                    obj    : A_Association ) is
    begin
        Boolean'Output( stream, obj /= null );
        if obj /= null then
            A_Association'Write( stream, obj );
        end if;
    end A_Association_Output;

    ----------------------------------------------------------------------------

    procedure A_Association_Read( stream : access Root_Stream_Type'Class;
                                  obj    : out A_Association ) is
    begin
        obj := new Association'Class'(Association'Class'Input( stream ));
    exception
        when e : Tag_Error =>
            declare
                msg : constant String := Exception_Message( e );
            begin
                raise Constraint_Error with "Found unknown association class " &
                      "<" & msg((Index( msg, ".", Backward ) + 1)..msg'Last) & ">";
            end;
    end A_Association_Read;

    ----------------------------------------------------------------------------

    procedure A_Association_Write( stream : access Root_Stream_Type'Class;
                                   obj    : A_Association ) is
    begin
         Association'Class'Output( stream, obj.all );
    end A_Association_Write;

    ----------------------------------------------------------------------------

    function Copy( src : A_Association ) return A_Association is
    begin
        return A_Association(Copy( A_Object(src) ));
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Association ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Associations;
