
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Containers.Indefinite_Hashed_Maps;
with Ada.Strings.Equal_Case_Insensitive;
with Ada.Strings.Hash_Case_Insensitive;
with Allegro.Files;                     use Allegro.Files;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support.Paths;                     use Support.Paths;

package body Audio_Players.Cache is

    PLAY_BUFFER_DEFAULT : constant :=  16 * 1024;
    PLAY_BUFFER_MAX     : constant := 128 * 1024;

    package Resource_Maps is new Ada.Containers.Indefinite_Hashed_Maps( String,
                                                                        A_Audio_Resource,
                                                                        Ada.Strings.Hash_Case_Insensitive,
                                                                        Ada.Strings.Equal_Case_Insensitive,
                                                                        "=" );
    use Resource_Maps;

    --==========================================================================

    function Create_Audio_Resource( filename : String ) return A_Audio_Resource is
        ext : constant String := To_Lower( Get_Extension( filename ) );
    begin
        if ext = "mp3" then
            return Create_Mp3_Resource( filename );
        elsif ext = "wav" or else ext = "voc" then
            return Create_Sample_Resource( filename );
        else
            raise RESOURCE_ERROR with "Unrecognized audio format <" & filename & '>';
        end if;
    end Create_Audio_Resource;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Audio_Resource;
                         filename : String ) is
    begin
        Limited_Object(this.all).Construct;
        this.resource := Resources.Load_Resource( filename, "audio" );
        if this.resource = null then
            raise RESOURCE_ERROR with "File not found <" & filename & '>';
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    function Create_Sound( this : access Audio_Resource ) return A_Sound is
        pragma Unreferenced( this );
    begin
        raise Constraint_Error with "Create_Sound must be implemented!";
        return null;
    end Create_Sound;

    ----------------------------------------------------------------------------

    procedure Dec_Refs( this : not null access Audio_Resource'Class ) is
    begin
        this.refs := this.refs - 1;
    end Dec_Refs;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Audio_Resource ) is
    begin
        Resources.Unload_Resource( this.resource );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Resource_File( this : not null access Audio_Resource'Class ) return A_Resource_File is
    begin
        return this.resource;
    end Get_Resource_File;

    ----------------------------------------------------------------------------

    procedure Inc_Refs( this : not null access Audio_Resource'Class ) is
    begin
        this.refs := this.refs + 1;
    end Inc_Refs;

    ----------------------------------------------------------------------------

    function Ref_Count( this : not null access Audio_Resource'Class ) return Natural is
    begin
        return this.refs;
    end Ref_Count;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Audio_Resource ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Create_Mp3_Resource( filename : String ) return A_Audio_Resource is
        this : A_Audio_Resource := new Mp3_Resource;
    begin
        this.Construct( filename );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Create_Mp3_Resource;

    ----------------------------------------------------------------------------

    overriding
    function Create_Sound( this : access Mp3_Resource ) return A_Sound is
    begin
        return Create_Mp3_Sound( A_Mp3_Resource(this) );  -- may raise exception
    end Create_Sound;

    --==========================================================================

    function Create_Sample_Resource( filename : String ) return A_Audio_Resource is
        this : A_Audio_Resource := new Sample_Resource;
    begin
        this.Construct( filename );
        return this;
    exception
        when others =>
            Delete( A_Audio_Resource(this) );
            raise;
    end Create_Sample_Resource;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Sample_Resource; filename : String ) is
        ext : constant String := To_Lower( Get_Extension( filename ) );
        pf  : A_Packfile;
    begin
        Audio_Resource(this.all).Construct( filename );
        pf := this.resource.Get_Packfile;
        if ext = "wav" then
            this.sample := Load_Wav_pf( pf );
        elsif ext = "voc" then
            this.sample := Load_Voc_pf( pf );
        end if;
        Pack_FClose0( pf );
        if this.sample = null then
            raise RESOURCE_ERROR with "File format error in <" & filename & '>';
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    function Create_Sound( this : access Sample_Resource ) return A_Sound is
    begin
        return Create_Sample_Sound( A_Sample_Resource(this) );  -- may raise exception
    end Create_Sound;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Sample_Resource ) is
    begin
        Destroy_Sample( this.sample );
        Audio_Resource(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Sample( this : not null access Sample_Resource'Class ) return A_Sample is
    begin
        return this.sample;
    end Get_Sample;

    --==========================================================================

    procedure Construct( this     : access Sound;
                         resource : not null A_Audio_Resource ) is
    begin
        Limited_Object(this.all).Construct;
        this.resource := resource;
    end Construct;

    ----------------------------------------------------------------------------

    function Get_Filename( this : not null access Sound'Class ) return String is
    begin
        return this.resource.Get_Resource_File.Get_Filename;
    end Get_Filename;

    ----------------------------------------------------------------------------

    function Is_Done( this : not null access Sound'Class ) return Boolean is
    begin
        return this.done;
    end Is_Done;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Sound ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    function Create_Mp3_Sound( resource : not null A_Mp3_Resource ) return A_Sound is
        this : A_Sound := new Mp3_Sound;
    begin
        Mp3_Sound(this.all).Construct( resource );
        return this;
    exception
        when others =>
            Delete( this );
            raise;
    end Create_Mp3_Sound;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Mp3_Sound;
                         resource : not null A_Mp3_Resource ) is
    begin
        Sound(this.all).Construct( A_Audio_Resource(resource) );
        this.mp3 := Create_mp3( this.resource.Get_Resource_File.Get_Address,
                                Integer(this.resource.Get_Resource_File.Size) );
        if this.mp3 = null then
            raise RESOURCE_ERROR with "File format error in <" &
                                      this.resource.Get_Resource_File.Get_Filename & '>';
        end if;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Mp3_Sound ) is
    begin
        Destroy_mp3( this.mp3 );
        Sound(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Play( this : access Mp3_Sound; looping : Boolean ) is
        stat    : Integer;
        bufsize : Integer := PLAY_BUFFER_DEFAULT;
    begin
        loop
            stat := Play_Ex_mp3( this.mp3, bufsize, vol => 255, pan => 127,
                                 speed => ALMP3_PLAY_SPEED_NORMAL,
                                 doloop => looping );
            bufsize := bufsize * 2;
            exit when stat /= ALMP3_PLAY_BUFFERTOOSMALL or else
                      bufsize > PLAY_BUFFER_MAX;
        end loop;
        if stat /= ALMP3_OK then
            pragma Debug(  Dbg( "Unable to play <" & this.Get_Filename & '>',
                           D_AUDIO, ERROR ) );
            this.done := True;
        end if;
    end Play;

    ----------------------------------------------------------------------------

    overriding
    procedure Poll( this : access Mp3_Sound ) is
        stat : Integer;
    begin
        stat := Poll_mp3( this.mp3 );
        case stat is
            when ALMP3_POLL_PLAYJUSTFINISHED |
                 ALMP3_POLL_NOTPLAYING =>
                this.done := True;
            when ALMP3_POLL_FRAMECORRUPT => null;
                pragma Debug( Dbg( "Audio_Player: Encountered corrupt frame in <" &
                                   this.Get_Filename & '>', D_AUDIO, ERROR ) );
            when ALMP3_POLL_INTERNALERROR =>
                 pragma Debug( Dbg( "Audio_Player: Unknown error playing <" &
                                    this.Get_Filename & '>', D_AUDIO, ERROR ) );
                 this.done := True;
            when others => null;
        end case;
    end Poll;

    ----------------------------------------------------------------------------

    overriding
    procedure Stop( this : access Mp3_Sound ) is
    begin
        if not this.done then
            Stop_mp3( this.mp3 );
            this.done := True;
        end if;
    end Stop;

    --==========================================================================

    function Create_Sample_Sound( resource : not null A_Sample_Resource ) return A_Sound is
        this : constant A_Sound := new Sample_Sound;
    begin
        this.Construct( A_Audio_Resource(resource) );
        return this;
    end Create_Sample_Sound;

    ----------------------------------------------------------------------------

    overriding
    procedure Play( this : access Sample_Sound; looping : Boolean ) is
    begin
        if not this.done then
            this.voice := Play_Sample( Sample_Resource(this.resource.all).Get_Sample,
                                       vol => 255, pan => 127, freq => 1000,
                                       loopit => Boolean'Pos( looping ) );
            if this.voice < 0 then
                this.done := True;
            end if;
        end if;
    end Play;

    ----------------------------------------------------------------------------

    overriding
    procedure Poll( this : access Sample_Sound ) is
    begin
        if not this.done then
            this.done := Voice_Check( this.voice ) /= Sample_Resource(this.resource.all).Get_Sample;
        end if;
    end Poll;

    ----------------------------------------------------------------------------

    overriding
    procedure Stop( this : access Sample_Sound ) is
    begin
        if not this.done then
            Deallocate_Voice( this.voice );
            this.done := True;
        end if;
    end Stop;

    --==========================================================================

    protected Audio_Cache is

        -- Raises an exception on error.
        procedure Load_Sound( filename : String; snd : out A_Sound );
        pragma Precondition( filename'Length > 0 );
        pragma Postcondition( snd /= null );

        procedure Unload_Sound( snd : in out A_Sound );
        pragma Precondition( snd /= null );
        pragma Postcondition( snd = null );

    private
        map : Resource_Maps.Map;
    end Audio_Cache;

    ----------------------------------------------------------------------------

    protected body Audio_Cache is

        procedure Load_Sound( filename : String; snd : out A_Sound ) is
            pos : constant Resource_Maps.Cursor := map.Find( filename );
            res : A_Audio_Resource;
        begin
            if Has_Element( pos ) then
                res := Element( pos );
            else
                res := Create_Audio_Resource( filename );  -- may raise exception
                map.Insert( filename, res );
            end if;

            snd := res.Create_Sound;    -- may raise exception

            -- this extra reference prevents the Audio_Cache from unloading it
            res.Inc_Refs;
        end Load_Sound;

        ------------------------------------------------------------------------

        procedure Unload_Sound( snd : in out A_Sound ) is
            filename : constant String := snd.Get_Filename;
            pos      : Resource_Maps.Cursor := map.Find( filename );
            res      : A_Audio_Resource;
        begin
            Delete( snd );
            if Has_Element( pos ) then
                res := Element( pos );
                res.Dec_Refs;
                if res.Ref_Count = 0 then
                    map.Delete( pos );
                    Delete( res );
                end if;
            else
                null;
                pragma Debug( Dbg( "Audio_Cache: " &
                                   "Reference count error for <" & filename & '>',
                                   D_AUDIO, ERROR ) );
            end if;
        end Unload_Sound;

    end Audio_Cache;

    --==========================================================================

    function Load_Sound( filename : String ) return A_Sound is
        sound : A_Sound;
    begin
        Audio_Cache.Load_Sound( filename, sound );
        return sound;
    end Load_Sound;

    ----------------------------------------------------------------------------

    procedure Unload_Sound( snd : in out A_Sound ) is
    begin
        if snd /= null then
            Audio_Cache.Unload_Sound( snd );
        end if;
    end Unload_Sound;

end Audio_Players.Cache;
