
with Events;                            use Events;
with Events.Corrals;                    use Events.Corrals;
with Events.Listeners;                  use Events.Listeners;
with Objects;                           use Objects;

private with Ada.Real_Time;
private with Ada.Strings.Unbounded;

package Audio_Players is

    -- Audio_Player objects listen for certain audio events and play sound
    -- effects or music on command. The player contains an internal task for
    -- playing audio in the background.
    type Audio_Player is new Limited_Object and Event_Listener with private;
    type A_Audio_Player is access all Audio_Player'Class;

    -- Creates a new Audio_Player that will listen for audio events in 'corral'.
    function Create_Audio_Player( corral : not null A_Corral ) return A_Audio_Player;
    pragma Postcondition( Create_Audio_Player'Result /= null );

    -- Starts the Audio_Player's internal task. This must be called before any
    -- audio can be played.
    procedure Start( this : not null access Audio_Player'Class );

    -- Stops the audio player's internals task. The destructor will
    -- automatically call Stop on deletion if the player hasn't already been
    -- stopped.
    procedure Stop( this : not null access Audio_Player'Class );

    -- Deletes the Audio_Player.
    procedure Delete( this : in out A_Audio_Player );
    pragma Postcondition( this = null );

private

    use Ada.Real_Time;
    use Ada.Strings.Unbounded;

    -- A Sound_Registry is a protected type that contains the set of sounds
    -- currently being played. Starting and Stopping sounds/music is done
    -- through the registry. The Polling_Task polls all sounds in the registry
    -- to keep the audio stream full.
    type Sound_Registry;
    type A_Sound_Registry is access all Sound_Registry;

    -- Deletes a Sound_Registry.
    procedure Delete( registry : in out A_Sound_Registry );
    pragma Postcondition( registry = null );

    -- Continuously polls audio files in a Sound_Registry to keep audio
    -- streaming.
    type Polling_Task;
    type A_Polling_Task is access all Polling_Task;

    -- Deletes a Polling_Task after it has stopped.
    procedure Delete( pollTask : in out A_Polling_Task );
    pragma Postcondition( pollTask = null );

    ----------------------------------------------------------------------------

    type Audio_Player is new Limited_Object and Event_Listener with
        record
            started   : Boolean := False;
            stopped   : Boolean := False;
            corral    : A_Corral := null;
            tickDelta : Time_Span := Time_Span_Zero;
            registry  : A_Sound_Registry := null;
            process   : A_Polling_Task := null;
            music     : Unbounded_String;
        end record;

    procedure Construct( this : access Audio_Player; corral : not null A_Corral );

    procedure Delete( this : in out Audio_Player );

    -- Handles Play_Music, Stop_Music, and Play_Sound events.
    procedure Handle_Event( this : access Audio_Player;
                            evt  : in out A_Event;
                            resp : out Response_Type );
    pragma Precondition( evt /= null );

    -- Called when a Play_Music event is received to begin playing a music track.
    -- Only one music track can play at a time so this will stop any previously
    -- playing music. Music tracks will loop.
    procedure Play_Music( this : not null access Audio_Player'Class; name : String );
    pragma Precondition( name'Length > 0 );

    -- Called when a Play_Sound event is received to begin playing a sound once
    -- through. Multiple sounds can play simultaneously.
    procedure Play_Sound( this : not null access Audio_Player'Class; name : String );
    pragma Precondition( name'Length > 0 );

    -- Called when a Stop_Music event is received. Any currently looping music
    -- will be stopped.
    procedure Stop_Music( this : not null access Audio_Player'Class );

end Audio_Players;
