
private with Ada.Containers.Ordered_Maps;
private with Locking_Objects;

generic
   type Element_Type is private;

   with function "="( left, right : Element_Type ) return Boolean is <>;

-- This generic package implements a thread-safe linked list that can be
-- modified by one thread while another is iterating. The list is locked at two
-- levels- the list level and the element level. The only restrictions are that
-- new elements can only be prepended or appended, and an element can't be
-- modified or removed from the list while it is being examined.
package Fully_Mutable_Lists is

    type Cursor is private;

    -- Closes the cursor, unlocking the list element it was examining.
    procedure Close( position : in out Cursor );

    -- Returns the value of the element the cursor is examining.
    function Element( position : Cursor ) return Element_Type;

    -- Returns True if the cursor is examining a valid element.
    function Has_Element( position : Cursor ) return Boolean;

    -- Traverses to the next element in the list. The caller will block until
    -- the next element is unlocked.
    procedure Next( position : in out Cursor );

    -- This is the fully mutable list type.
    type List is tagged limited private;

    -- Appends an element to the list.
    procedure Append( this : access List; element : Element_Type );

    -- Finds an element in the list, returning a cursor to it. If the element
    -- isn't in the list, an empty cursor will be returned.
    procedure Find( this : access List; element : Element_Type; position : out Cursor );

    -- Returns a cursor examining the first element in the list.
    function First( this : access List ) return Cursor;

    -- Returns True if the list is empty.
    function Is_Empty( this : access List ) return Boolean;

    -- Iterates forward over the list.
    procedure Iterate( this    : access List;
                       examine : access procedure( element : Element_Type ) );

    -- Iterates forward over the list with an early exit option. Set 'quit' to
    -- True in the 'examine' procedure to stop iterating.
    procedure Iterate_With_Quit( this    : access List;
                                 examine : access procedure( element : Element_Type;
                                                             quit    : in out Boolean ) );

    -- Returns the current size of the list. This should not be relied upon, as
    -- the list size may be changed by another thread immediately after return.
    function Length( this : access List ) return Natural;

    -- Prepends an element to the list.
    procedure Prepend( this : access List; element : Element_Type );

    -- It is possible that the position being removed is currently being
    -- examined. So- if the list contains pointers it is not safe to assume an
    -- object can be deleted after removing it from the list.
    procedure Remove( this : access List; position : in out Cursor );

    -- Raised when a cursor is used to access the list improperly.
    CONTAINER_ERROR : exception;

private

    use Locking_Objects;

    type Node is limited
        record
            element : Element_Type;
        end record;
    type A_Node is access all Node;

    procedure Delete( n : in out A_Node );

    package Element_Maps is new Ada.Containers.Ordered_Maps( Integer, A_Node, "<", "=" );

    type List is tagged limited
        record
            lock     : A_Locking_Object := new Locking_Object;
            least    : Integer := 0;
            greatest : Integer := 0;
            contents : Element_Maps.Map;
        end record;

    -- Result will be 0 / null if there isn't a following node.
    procedure Find_Next( this     : access List;
                         id       : Integer;
                         nextId   : out Integer;
                         nextNode : out A_Node );
    pragma Postcondition( nextId = 0 xor nextNode /= null );

    type Cursor is
        record
            list : access Fully_Mutable_Lists.List := null;
            id   : Integer := 0;
            node : A_Node := null;
        end record;

end Fully_Mutable_Lists;
