
with Ada.Streams;                       use Ada.Streams;
with Objects;                           use Objects;

package Maps is

    -- Maximum supported map dimensions
    MAX_WIDTH  : constant Positive;
    MAX_HEIGHT : constant Positive;

    ----------------------------------------------------------------------------

    -- A one dimensional array of tiles for a map layer
    type Layer_Data is array(Positive range <>) of Natural;

    -- A single map layer, where 'size' is the number of tiles in the layer.
    type Layer(size : Positive) is
        record
            physical : Boolean := False;     -- entities should clip to layer
            data     : Layer_Data(1..size);  -- tile ids (size = width * height)
        end record;
    type A_Layer is access all Layer;

    -- An array of map layers. All layers are expected to be the same size.
    type Layer_Array is array (Positive range <>) of A_Layer;
    type A_Layer_Array is access all Layer_Array;

    -- Returns a deep copy of a layer array.
    function Copy( src : A_Layer_Array ) return A_Layer_Array;

    -- Deep deletes a layer array.
    procedure Delete( la : in out A_Layer_Array );

    -- Moves a layer array reference. If 'dst' is not null it will be deleted.
    procedure Move( dst, src : in out A_Layer_Array );
    pragma Postcondition( src = null );
    pragma Postcondition( dst = src'Old );

    ----------------------------------------------------------------------------

    -- A Map_Object is layered, two-dimensional grid of tile ids used by Worlds.
    -- It's composed of a stack of Layers of identical dimensions. To get and
    -- set tile ids in the map, locations must be addressed by coordinates
    -- (layer, x, y). All coordinates start at 1. Layer 1 is the most background
    -- layer, ie: the furthest away in the z-order. The number of layers in a
    -- map is a constant, determined at construction by the concrete Map
    -- subclass.
    type Map_Object is abstract new Object with private;
    type A_Map is access all Map_Object'Class;

    -- Creates a new empty map with the given dimensions. The registered
    -- allocator will determine the concrete class of the map and how many
    -- layers it will have. An exception is raised if the dimensions are too large.
    function Create_Map( width, height : Positive ) return A_Map;
    pragma Postcondition( Create_Map'Result /= null );

    -- Returns the map's height in tiles.
    function Get_Height( this : not null access Map_Object'Class ) return Positive;
    pragma Postcondition( Get_Height'Result <= MAX_HEIGHT );

    -- Returns the number of layers in the map.
    function Get_Layers( this : not null access Map_Object'Class ) return Positive;

    -- Returns a pointer to the layer data owned by the map. Do not modify the
    -- contents.
    function Get_Layers_Data( this : not null access Map_Object'Class ) return A_Layer_Array;
    pragma Postcondition( Get_Layers_Data'Result /= null );

    -- Returns the id of a tile in the map. If the location is not on the map
    -- then 0 will be returned.
    function Get_Tile_Id( this  : not null access Map_Object'Class;
                          layer : Positive;
                          x, y  : Integer ) return Natural;

    -- Returns the width of the map in tiles.
    function Get_Width( this : not null access Map_Object'Class ) return Positive;
    pragma Postcondition( Get_Width'Result <= MAX_WIDTH );

    -- All concrete Map classes must implement this to be read from streams.
    function Object_Input( stream : access Root_Stream_Type'Class ) return Map_Object is abstract;

    -- Resizes the map. If the new size is larger in a dimension, the new space
    -- will be filled with tile id 0. If the new size is smaller in a dimension,
    -- the right and bottom edges will be clipped down to the new size. An
    -- exception will be raised if new dimensions are too large.
    procedure Resize( this   : not null access Map_Object'Class;
                      width,
                      height : Positive );

    -- Sets the tile id at a location in the map. If the location doesn't exist,
    -- nothing will happen.
    procedure Set_Tile( this  : not null access Map_Object'Class;
                        layer : Positive;
                        x, y  : Integer;
                        tile  : Natural );

    -- Returns the width of tiles in the map in pixels.
    function Tile_Width( this : not null access Map_Object'Class ) return Positive;

    -- Returns a deep copy of the map.
    function Copy( src : A_Map ) return A_Map;
    pragma Postcondition( Copy'Result /= src or else src = null );

    -- Deletes the map.
    procedure Delete( this : in out A_Map );
    pragma Postcondition( this = null );

private

    MAX_WIDTH  : constant Positive := 1024;
    MAX_HEIGHT : constant Positive := 1024;

    type Map_Object is abstract new Object with
        record
            width,
            height : Positive := 1;
            layers : A_Layer_Array := null;
        end record;

    procedure Adjust( this : access Map_Object );

    procedure Construct( this   : access Map_Object;
                         width,
                         height,
                         layers : Positive );

    procedure Delete( this : in out Map_Object );

    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Map_Object );
    for Map_Object'Read use Object_Read;

    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Map_Object );
    for Map_Object'Write use Object_Write;

    ----------------------------------------------------------------------------

    function A_Map_Input( stream : access Root_Stream_Type'Class ) return A_Map;
    for A_Map'Input use A_Map_Input;

    procedure A_Map_Output( stream : access Root_Stream_Type'Class; map : A_Map );
    for A_Map'Output use A_Map_Output;

    procedure A_Map_Read( stream : access Root_Stream_Type'Class; map : out A_Map );
    for A_Map'Read use A_Map_Read;

    procedure A_Map_Write( stream : access Root_Stream_Type'Class; map : A_Map );
    for A_Map'Write use A_Map_Write;

    ----------------------------------------------------------------------------

    -- Deletes the Layer.
    procedure Delete( lyr : in out A_Layer );

    function A_Layer_Input( stream : access Root_Stream_Type'Class ) return A_Layer;
    for A_Layer'Input use A_Layer_Input;

    procedure A_Layer_Output( stream : access Root_Stream_Type'Class; lyr : A_Layer );
    for A_Layer'Output use A_Layer_Output;

    ----------------------------------------------------------------------------

    function A_Layer_Array_Input( stream : access Root_Stream_Type'Class ) return A_Layer_Array;
    for A_Layer_Array'Input use A_Layer_Array_Input;

    procedure A_Layer_Array_Output( stream : access Root_Stream_Type'Class; la : A_Layer_Array );
    for A_Layer_Array'Output use A_Layer_Array_Output;

    ----------------------------------------------------------------------------

    -- An allocator function for creating a new Map.
    type Allocator is access function( width, height : Positive ) return A_Map;

    -- Registers the allocator used to create new Map instances.
    procedure Register_Allocator( allocate : not null Allocator );

end Maps;
