
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Containers.Indefinite_Ordered_Maps;
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Tags;                          use Ada.Tags;
with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;
with Debugging;                         use Debugging;
with Interfaces;                        use Interfaces;
with System;                            use System;

package body Objects is

    COUNT_INSTANCES : constant Boolean := False;

    type Mem_Info is
        record
            created : Integer := 0;
            deleted : Integer := 0;
        end record;

    package Instance_Count is new Ada.Containers.Indefinite_Ordered_Maps( String, Mem_Info, "<", "=" );
    use Instance_Count;

    ----------------------------------------------------------------------------

    protected Protected_Count is

        procedure Increment( class : String );

        procedure Decrement( class : String );

        procedure Display_Stats;

    private
        counts : Instance_Count.Map;
    end Protected_Count;

    ----------------------------------------------------------------------------

    protected body Protected_Count is

        procedure Increment( class : String ) is

             procedure Inc( class : String; info : in out Mem_Info ) is
                 pragma Unreferenced( class );
             begin
                 info.created := info.created + 1;
             end Inc;

            crsr : constant Instance_Count.Cursor := counts.Find( class );
        begin
            if not Has_Element( crsr ) then
                counts.Insert( class, (created=>1, deleted=>0) );
            else
                counts.Update_Element( crsr, Inc'Access );
            end if;
        end Increment;

        ------------------------------------------------------------------------

        procedure Decrement( class : String ) is

             procedure Dec( class : String; info : in out Mem_Info ) is
                 pragma Unreferenced( class );
             begin
                 info.deleted := info.deleted + 1;
             end Dec;

            crsr : constant Instance_Count.Cursor := counts.Find( class );
        begin
            if not Has_Element( crsr ) then
                counts.Insert( class, (created=>0, deleted=>1) );
            else
                counts.Update_Element( crsr, Dec'Access );
            end if;
        end Decrement;

        ------------------------------------------------------------------------

        procedure Display_Stats is

            procedure Display( pos : Instance_Count.Cursor ) is
                created : constant Integer := Element( pos ).created;
                deleted : constant Integer := Element( pos ).deleted;
            begin
                if created = deleted then
                    Dbg( Key( pos ) & " :" & created'Img & " created." );
                elsif created > deleted then
                    Dbg( Key( pos ) & " :" & created'Img & " created," &
                         Integer'Image( created - deleted ) & " leaked!" );
                else
                    Dbg( Key( pos ) & " :" & created'Img & " created /" &
                         deleted'Img & " deleted," & Integer'Image( deleted - created ) &
                         " unaccounted for!" );
                end if;
            end Display;

        begin
            Dbg( "--------------------------------------------------------------" );
            Dbg( "  Object Instantiation Stats" );
            Dbg( "--------------------------------------------------------------" );
            counts.Iterate( Display'Access );
            Dbg( "--------------------------------------------------------------" );
        end Display_Stats;

    end Protected_Count;

    ----------------------------------------------------------------------------

    procedure Display_Instance_Counts is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Display_Stats );
    end Display_Instance_Counts;

    --================================================================[ Object ]

    not overriding
    procedure Adjust( this : access Object ) is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Increment( this.Get_Class_Name( full => True ) ) );
    end Adjust;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this : access Object ) is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Increment( this.Get_Class_Name( full => True ) ) );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete( this : in out Object ) is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Decrement( this.Get_Class_Name( full => True ) ) );
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Class_Name( this : not null access Object'Class;
                             full : Boolean := False ) return String is
        tag : constant String := External_Tag( this.all'Tag );
        dot : constant Integer := Index( tag, ".", Backward );
    begin
        if not full and then dot > tag'First then
            return To_Lower( tag(dot+1..tag'Last) );
        end if;
        return To_Lower( tag );
    end Get_Class_Name;

    ----------------------------------------------------------------------------

    not overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Object ) is
        pragma Unreferenced( stream, obj );
    begin
        null;
    end Object_Read;

    ----------------------------------------------------------------------------

    not overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Object ) is
        pragma Unreferenced( stream, obj );
    begin
        null;
    end Object_Write;

    ----------------------------------------------------------------------------

    not overriding
    function To_String( this : access Object ) return String is
        function To_Unsigned_32 is new Ada.Unchecked_Conversion( Address, Unsigned_32 );
    begin
        return "<" & this.Get_Class_Name & " @" &
                     Unsigned_32'Image( To_Unsigned_32( this.all'Address ) ) &
               ">";
    end To_String;

    --==============================================================[ A_Object ]

    procedure Free is new Ada.Unchecked_Deallocation( Object'Class, A_Object );

    ----------------------------------------------------------------------------

    function Copy( src : A_Object ) return A_Object is
        dest : A_Object := null;
    begin
        if src /= null then
            dest := new Object'Class'(src.all);
            begin
                dest.Adjust;
            exception
                when COPY_NOT_ALLOWED =>
                    Free( dest );
                    raise;
            end;
        end if;
        return dest;
    end Copy;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Object ) is
    begin
        if this /= null then
            this.Delete;
                Free( this );
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    function "&"( left : A_Object; right : String ) return String is
    begin
        if left = null then
            return "null" & right;
        else
            return left.To_String & right;
        end if;
    end "&";

    ----------------------------------------------------------------------------

    function "&"( left : String; right : A_Object ) return String is
    begin
        if right = null then
            return left & "null";
        else
            return left & right.To_String;
        end if;
    end "&";

    --========================================================[ Limited_Object ]

    not overriding
    procedure Construct( this : access Limited_Object ) is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Increment( this.Get_Class_Name( full => True ) ) );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Delete( this : in out Limited_Object ) is
    begin
        null;
        pragma Debug( COUNT_INSTANCES, Protected_Count.Decrement( this.Get_Class_Name( full => True ) ) );
    end Delete;

    ----------------------------------------------------------------------------

    function Get_Class_Name( this : not null access Limited_Object'Class;
                             full : Boolean := False ) return String is
        tag : constant String := External_Tag( this.all'Tag );
        dot : constant Integer := Index( tag, ".", Backward );
    begin
        if not full and then dot > tag'First then
            return To_Lower( tag(dot+1..tag'Last) );
        end if;
        return To_Lower( tag );
    end Get_Class_Name;

    ----------------------------------------------------------------------------

    not overriding
    function To_String( this : access Limited_Object ) return String is
        function To_Unsigned_32 is new Ada.Unchecked_Conversion( Address, Unsigned_32 );
    begin
        return "<" & this.Get_Class_Name & " @" &
                     Unsigned_32'Image( To_Unsigned_32( this.all'Address ) ) &
               ">";
    end To_String;

    --======================================================[ A_Limited_Object ]

    procedure Free is new Ada.Unchecked_Deallocation( Limited_Object'Class, A_Limited_Object );

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Limited_Object ) is
    begin
        if this /= null then
            this.Delete;
                Free( this );
        end if;
    end Delete;

    ----------------------------------------------------------------------------

    function "&"( left : A_Limited_Object; right : String ) return String is
    begin
        if left = null then
            return "null" & right;
        else
            return left.To_String & right;
        end if;
    end "&";

    ----------------------------------------------------------------------------

    function "&"( left : String; right : A_Limited_Object ) return String is
    begin
        if right = null then
            return left & "null";
        else
            return left & right.To_String;
        end if;
    end "&";

end Objects;
