
with Entities;                          use Entities;
with Objects;                           use Objects;
with Physics.Clip_Maps;                 use Physics.Clip_Maps;

limited with Physics.Managers;

private with Ada.Containers.Ordered_Maps;

package Physics.Bodies is

    -- A physical representation of an entity, used by a Phsyics_Manager.
    type Corpus is new Object with private;
    type A_Corpus is access all Corpus'Class;

    -- Creates a new corpus representing entity 'id', for use by 'manager'.
    function Create_Corpus( id       : Entity_Id;
                            width,
                            height   : Natural;
                            physical : Boolean;
                            clipped  : Boolean;
                            manager  : access Physics.Managers.Physics_Manager'Class
                          ) return A_Corpus;

    -- Performs collision detection and returns True if this corpus is touching
     -- 'that' corpus. The objects are not modified.
    function Detect_Collision( this : not null access Corpus'Class;
                               that : not null A_Corpus ) return Boolean;

    -- Returns the entity id the corpus represents.
    function Get_Id( this : not null access Corpus'Class ) return Entity_Id;

    -- Returns True if 'that' corpus is in this corpus' touch list. No collision
    -- detection is performed.
    function Is_Touching( this : not null access Corpus'Class;
                          that : not null A_Corpus ) return Boolean;

    -- Marks the contact state between this corpus and 'that'. The corpus' touch
    -- list will be updated. Set 'inContact' to true if the two corps are in
    -- contact with each other. Corpus 'that' is not modified.
    procedure Mark_Contact( this      : not null access Corpus'Class;
                            that      : not null A_Corpus;
                            inContact : Boolean );

    -- Moves the entity to the given location. An Entity_Moved event is queued.
    -- Clipping is performed so an Entity_Grounded event may also be queued.
    procedure Move_To( this : not null access Corpus'Class;
                       map  : not null A_Clip_Map;
                       x, y : Float );

    -- Sets the acceleration in the X axis.
    procedure Set_AX( this : not null access Corpus'Class; ax : Float );

    -- Sets the acceleration in the Y axis.
    procedure Set_AY( this : not null access Corpus'Class; ay : Float );

    -- Sets the clipped attribute of the Corpus. Clipped corps are bounded by
    -- the wall tiles of the map.
    procedure Set_Clipped( this    : not null access Corpus'Class;
                           clipped : Boolean );

    -- Sets the Corpus' location. No event is queued.
    procedure Set_Location( this : not null access Corpus'Class;
                            map  : not null A_Clip_Map;
                            x, y : Float );

    -- Sets the physical attribute of the Corpus. Physical corps are affected by
    -- the forces of gravity and friction.
    procedure Set_Physical( this     : not null access Corpus'Class;
                            physical : Boolean );

    -- Sets the physical size of the Corpus. The Corpus is clipped to 'map'
    -- immediately. An Entity_Resized event is queued.
    procedure Set_Size( this   : not null access Corpus'Class;
                        map    : not null A_Clip_Map;
                        width,
                        height : Natural );

    -- Sets the target velocity in the X axis to accelerate toward. This does
    -- not have an immediate effect on the velocity.
    procedure Set_Target_VX( this : not null access Corpus'Class; vx : Float );

    -- Sets the target velocity in the Y axis to accelerate toward. This does
    -- not have an immediate effect on the velocity.
    procedure Set_Target_VY( this : not null access Corpus'Class; vy : Float );

    -- Sets the Corpus' velocity.
    procedure Set_Velocity( this : not null access Corpus'Class; vx, vy : Float );

    -- Ticks the Corpus' changes in velocity and position and performs clipping.
    -- If 'collide' is returned True then the Corpus location changed and
    -- collision detection needs to be performed on it.
    procedure Tick( this    : not null access Corpus'Class;
                    map     : not null A_Clip_Map;
                    dt      : Float;
                    collide : out Boolean );

    -- Returns a string representation of the Corpus.
    function To_String( this : access Corpus'Class ) return String;

    -- Deletes the Corpus.
    procedure Delete( this : in out A_Corpus );
    pragma Postcondition( this = null );

private

    package Corps_Maps is new Ada.Containers.Ordered_Maps( Entity_Id, A_Corpus, "<", "=" );

    type Corpus is new Object with
        record
            manager    : access Physics.Managers.Physics_Manager'Class := null;
            eid        : Entity_Id := INVALID_ID;
            width,
            height     : Natural := 0;
            width_2,
            height_2   : Natural := 0;
            physical   : Boolean := True;    -- apply gravity and friction
            clipped    : Boolean := True;    -- collide with the world
            collide    : Boolean := True;    -- need to check for collisions with other corps

            x, y,
            oldX, oldY : Float := 0.0;       -- old x,y used only during a Tick
            vx, vy,
            tvx, tvy,
            ax, ay     : Float := 0.0;
            standing   : Boolean := False;
            touching   : Corps_Maps.Map;
        end record;

    -- Updates the Corpus' velocity by applying all acting forces, including
    -- gravity, friction, and target acceleration.
    procedure Apply_Forces( this : not null access Corpus'Class; dt : Float );

    -- Clips the Corpus to the solid tile types in 'map', updating the Corpus
    -- location and zeroing its velocity as necessary. Collision-related events
    -- will be queued as they are detected.
    procedure Clip( this : not null access Corpus'Class; map : not null A_Clip_Map );

    procedure Construct( this     : access Corpus;
                         id       : Entity_Id;
                         width,
                         height   : Natural;
                         physical : Boolean;
                         clipped  : Boolean;
                         manager  : access Physics.Managers.Physics_Manager'Class );

    procedure Delete( this : in out Corpus );

end Physics.Bodies;
