
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Real_Time;                     use Ada.Real_Time;
with Debugging;                         use Debugging;
with Directions;                        use Directions;
with Engine_Debugging;                  use Engine_Debugging;
with Preferences;                       use Preferences;
with Support;                           use Support;
with Values;                            use Values;

package body Physics.Managers is

    use Corpus_Map;

    ----------------------------------------------------------------------------

    function Create_Physics( corral : not null A_Corral ) return A_Physics is
        this : constant A_Physics := new Physics_Manager;
    begin
        this.Construct( corral );
        return this;
    end Create_Physics;

    ----------------------------------------------------------------------------

    procedure Clear_Corps( this : not null access Physics_Manager'Class ) is

        procedure Examine( pos : Corpus_Map.Cursor ) is
            corp : A_Corpus := Element( pos );
        begin
            Delete( corp );
        end Examine;

    begin
        this.corps.Iterate( Examine'Access );
        this.corps.Clear;
    end Clear_Corps;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this   : access Physics_Manager;
                         corral : not null A_Corral ) is
    begin
        Object(this.all).Construct;
        this.corral := corral;
        -- act on commands
        this.corral.Add_Listener( A_Event_Listener(this), "Accelerate" );
        this.corral.Add_Listener( A_Event_Listener(this), "Move_Entity" );
        this.corral.Add_Listener( A_Event_Listener(this), "Resize_Entity" );
        this.corral.Add_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );

        -- respond to events that have occurred
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Created" );
        this.corral.Add_Listener( A_Event_Listener(this), "Entity_Deleted" );
        this.corral.Add_Listener( A_Event_Listener(this), "New_World" );
        this.corral.Add_Listener( A_Event_Listener(this), "Tile_Changed" );
        this.corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Physics_Manager ) is
    begin
        this.Clear_Corps;
        Delete( this.map );
        this.corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Tile_Changed" );
        this.corral.Remove_Listener( A_Event_Listener(this), "New_World" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Deleted" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Entity_Created" );

        this.corral.Remove_Listener( A_Event_Listener(this), "Set_Entity_Attribute" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Resize_Entity" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Move_Entity" );
        this.corral.Remove_Listener( A_Event_Listener(this), "Accelerate" );
        Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    function Find( this : not null access Physics_Manager'Class;
                   id   : Entity_Id ) return A_Corpus is
        pos : constant Corpus_Map.Cursor := this.corps.Find( id );
    begin
        if Has_Element( pos ) then
            return Element( pos );
        end if;
        return null;
    end Find;

    ----------------------------------------------------------------------------

    function Friction_X( this : not null access Physics_Manager'Class ) return Float is
    begin
        return this.fricX;
    end Friction_X;

    ----------------------------------------------------------------------------

    function Friction_Y( this : not null access Physics_Manager'Class ) return Float is
    begin
        return this.fricY;
    end Friction_Y;

    ----------------------------------------------------------------------------

    overriding
    function Get_Process_Name( this : access Physics_Manager ) return String is
        pragma Unreferenced( this );
    begin
        return "Physics Manager";
    end Get_Process_Name;

    ----------------------------------------------------------------------------

    function Gravity( this : not null access Physics_Manager'Class ) return Float is
    begin
        return this.grav;
    end Gravity;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Accelerate_Event ) is
        corp : constant A_Corpus := this.Find( evt.Get_Id );
    begin
        -- convert velocity and acceleration values into proper axis using the
        -- direction of the acceleration. if velocity is zero, a non-zero
        -- acceleration value will cause deceleration to zero at the given rate.
        if corp /= null then
            case evt.Get_Direction is
                when Left =>
                    corp.Set_Target_VX( -evt.Get_Velocity );
                    corp.Set_AX( evt.Get_Acceleration );
                when Right =>
                    corp.Set_Target_VX( evt.Get_Velocity );
                    corp.Set_AX( evt.Get_Acceleration );
                when Up =>
                    corp.Set_Target_VY( -evt.Get_Velocity );
                    corp.Set_AY( evt.Get_Acceleration );
                when Down =>
                    corp.Set_Target_VY( evt.Get_Velocity );
                    corp.Set_AY( evt.Get_Acceleration );
            end case;
            pragma Debug( Dbg( "Physics_Manager.Accelerate( " &
                               Entity_Id'Image( evt.Get_Id ) & ", " &
                               Cardinal_Direction'Image( evt.Get_Direction ) & ", " &
                               Image( evt.Get_Velocity ) & ", " &
                               Image( evt.Get_Acceleration ) & " )",
                               D_PHYSICS, INFO ) );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Entity_Created_Event ) is
        corp : A_Corpus;
    begin
        if this.map /= null then
            corp := Create_Corpus( evt.Get_Id, evt.Get_Width, evt.Get_Height,
                                   evt.Is_Physical, evt.Is_Clipped, this );
            corp.Set_Velocity( evt.Get_XV, evt.Get_YV );
            corp.Set_Location( this.map, evt.Get_X, evt.Get_Y );
            this.corps.Insert( evt.Get_Id, corp );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Entity_Deleted_Event ) is
        pos  : Corpus_Map.Cursor := this.corps.Find( evt.Get_Id );
        corp : A_Corpus;
    begin
        if Has_Element( pos ) then
            corp := Element( pos );
            this.corps.Delete( pos );
            Delete( corp );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Move_Entity_Event ) is
        corp : constant A_Corpus := this.Find( evt.Get_Id );
    begin
        if corp /= null then
            corp.Move_To( this.map, evt.Get_X, evt.Get_Y );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_New_World_Event ) is
    begin
        Delete( this.map );
        this.Clear_Corps;
        this.map := Create_Clip_Map( evt.Get_Width, evt.Get_Height,
                                     evt.Get_Tile_Width, evt.Get_Layers,
                                     evt.Get_Library_Name );
    exception
        when e : others =>
            -- Exception while creating clip map
            pragma Debug( Dbg( "Physics_Manager: Error handling New_World_Event: " &
                               Exception_Name( e ) & " - " & Exception_Message( e ),
                               D_PHYSICS, ERROR ) );
            this.map := null;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Resize_Entity_Event ) is
        corp : constant A_Corpus := this.Find( evt.Get_Id );
    begin
        if corp /= null then
            corp.Set_Size( this.map, evt.Get_Width, evt.Get_Height );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Set_Entity_Attribute_Event ) is
        attr : constant String := To_Lower( evt.Get_Attribute );
        corp : A_Corpus;
        val  : A_Value;
    begin
        if attr = "physical" then
            corp := this.Find( evt.Get_Id );
            evt.Copy_Value( val );
            if val /= null then
                corp.Set_Physical( val.As_Boolean );
            end if;
            Delete( val );
        elsif attr = "clipped" then
            corp := this.Find( evt.Get_Id );
            evt.Copy_Value( val );
            if val /= null then
                corp.Set_Clipped( val.As_Boolean );
            end if;
            Delete( val );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_Tile_Changed_Event ) is
    begin
        if this.map /= null then
            this.map.Set_Tile( evt.Get_Tile_ID, evt.Get_X, evt.Get_Y, evt.Get_Layer );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    procedure Handle( this : not null access Physics_Manager'Class;
                      evt  : not null A_World_Property_Changed_Event ) is
    begin
        if evt.Get_Property_Name = "domain" then
            this.grav  := Constrain( Get_Pref( "physics", evt.Get_Value & ".gravity" ), 0.0, 10000.0 );
            this.fricX := Constrain( Get_Pref( "physics", evt.Get_Value & ".friction_x" ), 0.0, 1.0 );
            this.fricY := Constrain( Get_Pref( "physics", evt.Get_Value & ".friction_y" ), 0.0, 1.0 );
            this.maxVX := Constrain( Get_Pref( "physics", evt.Get_Value & ".max_speed_x" ), 0.0, 500.0 );
            this.maxVY := Constrain( Get_Pref( "physics", evt.Get_Value & ".max_speed_y" ), 0.0, 500.0 );
        end if;
    end Handle;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Physics_Manager;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        -- these are roughly ordered according to their frequency; most frequent
        -- first to minimize comparisons.

        if evt.Get_Id = ACCELERATE_ID then
            this.Handle( A_Accelerate_Event(evt) );

        elsif evt.Get_Id = MOVE_ENTITY_ID then
            this.Handle( A_Move_Entity_Event(evt) );

        elsif evt.Get_Id = ENTITY_CREATED_ID then
            this.Handle( A_Entity_Created_Event(evt) );

        elsif evt.Get_Id = ENTITY_DELETED_ID then
            this.Handle( A_Entity_Deleted_Event(evt) );

        elsif evt.Get_Id = TILE_CHANGED_ID then
            this.Handle( A_Tile_Changed_Event(evt) );

        elsif evt.Get_Id = SET_ENTITY_ATTRIBUTE_ID then
            this.Handle( A_Set_Entity_Attribute_Event(evt) );

        elsif evt.Get_Id = RESIZE_ENTITY_ID then
            this.Handle( A_Resize_Entity_Event(evt) );

        elsif evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
            this.Handle( A_World_Property_Changed_Event(evt) );

        elsif evt.Get_Id = NEW_WORLD_ID then
            this.Handle( A_New_World_Event(evt) );

        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    function Max_VX( this : not null access Physics_Manager'Class ) return Float is
    begin
        return this.maxVX;
    end Max_VX;

    ----------------------------------------------------------------------------

    function Max_VY( this : not null access Physics_Manager'Class ) return Float is
    begin
        return this.maxVY;
    end Max_VY;

    ----------------------------------------------------------------------------

    overriding
    procedure Tick( this : access Physics_Manager; time : Tick_Time ) is

        movers : Corpus_Lists.List;

        ------------------------------------------------------------------------

        procedure Tick_All( dt : Float ) is

            --------------------------------------------------------------------

            procedure Tick_One( pos : Corpus_Map.Cursor ) is
                corp    : constant A_Corpus := Element( pos );
                collide : Boolean;
            begin
                corp.Tick( this.map, dt, collide );
                if collide then
                    movers.Append( corp );
                end if;
            end Tick_One;

            --------------------------------------------------------------------

        begin
            this.corps.Iterate( Tick_One'Access );
        end Tick_All;

        ------------------------------------------------------------------------

        procedure Detect_Collisions is
            first : Corpus_Map.Cursor := Next( this.corps.First );

            --------------------------------------------------------------------

            procedure Collide_Each( pos1 : Corpus_Lists.Cursor ) is
                corp1       : constant A_Corpus := Corpus_Lists.Element( pos1 );
                corp2       : A_Corpus;
                pos2        : Corpus_Map.Cursor := first;
                wasTouching,
                nowTouching : Boolean;
            begin
                while Has_Element( pos2 ) loop
                    corp2 := Element( pos2 );
                    if corp2 /= corp1 then
                        wasTouching := corp1.Is_Touching( corp2 );
                        nowTouching := corp1.Detect_Collision( corp2 );
                        corp1.Mark_Contact( corp2, inContact => nowTouching );
                        corp2.Mark_Contact( corp1, inContact => nowTouching );
                        if not wasTouching and then nowTouching then
                            Queue_Entities_Collided( corp1.Get_Id, Key( pos2 ) );
                        elsif wasTouching and then not nowTouching then
                            Queue_Entities_Separated( corp1.Get_Id, Key( pos2 ) );
                        end if;
                    end if;
                    pos2 := Next( pos2 );
                end loop;
                first := Next( first );
            end Collide_Each;

            --------------------------------------------------------------------

        begin
            movers.Iterate( Collide_Each'Access );
        end Detect_Collisions;

        ------------------------------------------------------------------------

    begin
        Tick_All( Float(To_Duration( time.elapsed )) );
        Detect_Collisions;
    end Tick;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Physics ) is
    begin
        Delete( A_Object(this) );
    end Delete;

end Physics.Managers;
