
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Ada.Unchecked_Deallocation;
with Applications;                      use Applications;
with Applications.Gui.Games;            use Applications.Gui.Games;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support;                           use Support;
with Support.Real_Time;                 use Support.Real_Time;

package body Processes.Managers is

    function Create_Process_Manager( name     : String;
                                     hertz    : Natural := 0;
                                     minHertz : Natural := 0
                                   ) return A_Process_Manager is
        this : constant A_Process_Manager := new Process_Manager;
    begin
        this.Construct( name, hertz, minHertz );
        return this;
    end Create_Process_Manager;

    ----------------------------------------------------------------------------

    procedure Attach( this : not null access Process_Manager'Class;
                      proc : not null A_Process ) is
        op : A_Operation;
    begin
        op := Create_Attach_Operation( proc );
        this.lock.Lock;
        this.operations.Append( op );
        this.lock.Unlock;
    end Attach;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this     : access Process_Manager;
                         name     : String;
                         hertz,
                         minHertz : Natural ) is
    begin
        Limited_Object(this.all).Construct;
        this.lock := new Locking_Object;
        this.name := To_Unbounded_String( name );
        this.ticker := new Ticker_Task;
        this.ticker.Init( A_Process_Manager(this), hertz, minHertz );
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Process_Manager ) is

        procedure Delete_Op( pos : Operation_Lists.Cursor ) is
            op : A_Operation;
        begin
            op := Operation_Lists.Element( pos );
            Delete( op );
        end Delete_Op;

    begin
        this.Stop;
        this.lock.Lock;
        this.operations.Iterate( Delete_Op'Access );
        this.operations.Clear;
        Delete( this.lock );
        Delete( this.ticker );
        Limited_Object(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    procedure Detach( this       : not null access Process_Manager'Class;
                      proc       : not null A_Process;
                      destructor : A_Destructor := null ) is
        op : A_Operation;
    begin
        op := Create_Detach_Operation( proc, destructor );
        this.lock.Lock;
        this.operations.Append( op );
        this.lock.Unlock;
    end Detach;

    ----------------------------------------------------------------------------

    function Get_Name( this : not null access Process_Manager'Class ) return String is
    begin
        return To_String( this.name );
    end Get_Name;

    ----------------------------------------------------------------------------

    function Get_Rate( this : not null access Process_Manager'Class ) return Natural is
        result : Natural := 0;
    begin
        this.lock.Lock;
        if this.trackRate then
            result := this.rate;
        end if;
        this.lock.Unlock;
        return result;
    end Get_Rate;

    ----------------------------------------------------------------------------

    procedure Pause( this   : not null access Process_Manager'Class;
                     proc   : not null A_Process;
                     paused : Boolean )  is
        op : A_Operation;
    begin
        if paused then
            op := Create_Pause_Operation( proc );
        else
            op := Create_Resume_Operation( proc );
        end if;
        this.lock.Lock;
        this.operations.Append( op );
        this.lock.Unlock;
    end Pause;

    ----------------------------------------------------------------------------

    procedure Start( this : not null access Process_Manager'Class ) is
    begin
        this.lock.Lock;
        if not this.started and then not this.stopped then
            this.started := True;
            this.ticker.Start;
        end if;
        this.lock.Unlock;
    end Start;

    ----------------------------------------------------------------------------

    procedure Stop( this : not null access Process_Manager'Class ) is
    begin
        this.lock.Lock;
        if not this.stopped then
            this.stopped := True;

            -- the ticker task needs the lock, so we must release it before
            -- attempting to rendezvous with ticker to prevent a deadlock.
            this.lock.Unlock;

            this.ticker.Stop;

            for count in 1..10 loop
                if Is_Terminated( this.ticker'Identity ) then
                    exit;
                end if;
                delay 0.1;
            end loop;
            if not Is_Terminated( this.ticker'Identity ) then
                null;
                pragma Debug( Dbg( "Process_Manager.Stop: Ticker_Task failed to stop",
                                   D_INIT, ERROR ) );
            end if;
        else
            this.lock.Unlock;
        end if;
    end Stop;

    ----------------------------------------------------------------------------

    procedure Track_Rate( this    : not null access Process_Manager'Class;
                          enabled : Boolean ) is
    begin
        this.lock.Lock;
        this.trackRate := enabled;
        this.lock.Unlock;
    end Track_Rate;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Process_Manager ) is
    begin
        Delete( A_Limited_Object(this) );
    end Delete;

    --==========================================================================

    procedure Delete( exec : in out A_Execution ) is
        procedure Free is new Ada.Unchecked_Deallocation( Execution, A_Execution );
    begin
        Free( exec );
    end Delete;

    ----------------------------------------------------------------------------

    function Eq( l, r : A_Execution ) return Boolean is
    begin
        if l = null then
            return r = null;
        elsif r = null then
            return False;
        else
            return l.process = r.process;
        end if;
    end Eq;

    ----------------------------------------------------------------------------

    procedure Delete( this : in out A_Operation ) is
        procedure Free is new Ada.Unchecked_Deallocation( Operation'Class, A_Operation );
    begin
        Free( this );
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Attach_Operation;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process   => this.process,
                                             firstTick => Clock,
                                             lastTick  => Clock,
                                             others    => <>);
    begin
        if not execList.Contains( exec ) then
            execList.Append( exec );
            pragma Debug( Dbg( "Attached process <" &
                               exec.process.Get_Process_Name & '>',
                               D_PROCS, INFO ) );
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Detach_Operation;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            execList.Delete( crsr );
            pragma Debug( Dbg( "Detached process <" &
                               exec.process.Get_Process_Name & '>',
                               D_PROCS, INFO ) );
            if this.destroy /= null then
                this.destroy( exec.process );
            end if;
        end if;
        Delete( exec );
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Pause_Operation;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            exec.paused := True;
            exec.pauseTime := Clock;
            pragma Debug( Dbg( "Paused process <" &
                               exec.process.Get_Process_Name & '>',
                               D_PROCS, INFO ) );
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    overriding
    procedure Execute( this     : access Resume_Operation;
                       execList : in out Execution_Lists.List ) is
        exec : A_Execution := new Execution'(process => this.process, others => <>);
        crsr : Execution_Lists.Cursor;
    begin
        crsr := execList.Find( exec );
        if Execution_Lists.Has_Element( crsr ) then
            Delete( exec );
            exec := Execution_Lists.Element( crsr );
            pragma Debug( Dbg( "Resuming process <" &
                               exec.process.Get_Process_Name & '>',
                               D_PROCS, INFO ) );
            if exec.paused then
                exec.paused := False;
                exec.lastTick := Clock - (exec.pauseTime - exec.lastTick);
            end if;
        else
            Delete( exec );
        end if;
    end Execute;

    ----------------------------------------------------------------------------

    function Create_Attach_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Attach_Operation;
    begin
        op.process := process;
        return op;
    end Create_Attach_Operation;

    ----------------------------------------------------------------------------

    function Create_Detach_Operation( process    : not null A_Process;
                                      destructor : A_Destructor ) return A_Operation is
        op : constant A_Operation := new Detach_Operation;
    begin
        op.process := process;
        Detach_Operation(op.all).destroy := destructor;
        return op;
    end Create_Detach_Operation;

    ----------------------------------------------------------------------------

    function Create_Pause_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Pause_Operation;
    begin
        op.process := process;
        return op;
    end Create_Pause_Operation;

    ----------------------------------------------------------------------------

    function Create_Resume_Operation( process : not null A_Process ) return A_Operation is
        op : constant A_Operation := new Resume_Operation;
    begin
        op.process := process;
        return op;
    end Create_Resume_Operation;

    --==========================================================================

    task body Ticker_Task is
        this : A_Process_Manager;
        fps  : aliased Frame_Timer;   -- for frame rate counter

        ------------------------------------------------------------------------

#if DEBUG then
        nextUsageUpdate : Time := Clock;
        usageTotal      : Natural := 0;
        usageCount      : Natural := 0;

        -- Track the CPU time used by the process manager. The percentage used
        -- can be calculated by checking the difference between the requested
        -- time between frames and the amount of time remaining for sleep after
        -- ticking all processes in the frame. This code is debugging only so it
        -- doesn't slow down the release version. The average CPU use each
        -- second is sent to debugging output and can be seen with INFO level
        -- output for the PROCS system.
        procedure Track_CPU_Usage( tickDelta : Time_Span;
                                   nextTick  : Time ) is
            now       : constant Time := Clock;
            remaining : Time_Span;
            usageNow  : Natural := 0;
        begin
            if now > nextTick then
                usageNow := 100;
            else
                remaining := nextTick - now;
                usageNow := 100 - Integer(Float(100.0) * Float(To_Duration( remaining ) / To_Duration( tickDelta )));
            end if;
            usageTotal := usageTotal + usageNow;
            usageCount := usageCount + 1;

            if now > nextUsageUpdate then
                Dbg( "PMan(" & this.Get_Name & ") CPU:" &
                     Integer'Image( usageTotal / usageCount ) & "%",
                     D_PROCS, INFO );
                usageTotal := 0;
                usageCount := 0;
                nextUsageUpdate := now + Seconds( 1 );
            end if;
        end Track_CPU_Usage;
#end if;

        ------------------------------------------------------------------------

        procedure Tick( minDelta : Time_Span;
                        execList : in out Execution_Lists.List ) is
            now : constant Time := Clock;

            --------------------------------------------------------------------

            procedure Exec_Op( pos : Operation_Lists.Cursor ) is
                op : constant A_Operation := Operation_Lists.Element( pos );
            begin
                op.Execute( execList );
            end Exec_Op;

            --------------------------------------------------------------------

            procedure Tick_Proc( pos : Execution_Lists.Cursor ) is
                exec : constant A_Execution := Execution_Lists.Element( pos );
                dt   : Time_Span;
            begin
                if not exec.paused then
                    dt := now - exec.lastTick;
                    if minDelta > Time_Span_Zero and then dt > minDelta then
                        dt := minDelta;
                    end if;
                    exec.process.Tick( Tick_Time'(total => now - exec.firstTick, elapsed => dt) );
                    exec.lastTick := now;
                end if;
            exception
                when e : others =>
                    -- the process threw an unhandled exception. terminate then
                    -- application.
                    Terminate_Application( "Exception in process <" &
                                           exec.process.Get_Process_Name & ">: " &
                                           Exception_Message( e ) );
            end Tick_Proc;

            --------------------------------------------------------------------

        begin
            this.lock.Lock;
            this.operations.Iterate( Exec_Op'Access );
            this.operations.Clear;
            if this.trackRate then
                Tick( fps );
                if Is_Updated( fps ) then
                    this.rate := Rate( fps'Access );
                end if;
            end if;
            this.lock.Unlock;

            execList.Iterate( Tick_Proc'Access );
        end Tick;

        ------------------------------------------------------------------------

        stopped   : Boolean := False;
        nextTick  : Time;
        tickDelta,
        minDelta  : Time_Span := Time_Span_Zero;
        execList  : Execution_Lists.List;
    begin
        pragma Debug( Dbg( "Processes.Managers.Ticker_Task created", D_INIT, INFO ) );

        accept Init( pman : A_Process_Manager; hertz, minHertz : Natural ) do
            this := pman;
            if hertz > 0 then
                tickDelta := Seconds( 1 ) / hertz;
            end if;
            if minHertz > 0 then
                minDelta := Seconds( 1 ) / minHertz;
            end if;
            pragma Debug( Dbg( "Processes.Managers.Ticker_Task initialized", D_INIT, INFO ) );
        end Init;

        select
            accept Start do
                null;
                pragma Debug( Dbg( "Processes.Managers.Ticker_Task starting", D_INIT, INFO ) );
            end Start;
        or
            accept Stop do
                pragma Debug( Dbg( "Processes.Managers.Ticker_Task stopping", D_INIT, INFO ) );
                stopped := True;
            end Stop;
        end select;

        nextTick := Clock + tickDelta;
        while not stopped loop
            Tick( minDelta, execList );
#if DEBUG then
            Track_CPU_Usage( tickDelta, nextTick );
#end if;

            select
                accept Stop do
                    pragma Debug( Dbg( "Processes.Managers.Ticker_Task stopping", D_INIT, INFO ) );
                    stopped := True;
                end Stop;
            or
                delay until nextTick;
            end select;

            nextTick := nextTick + tickDelta;
        end loop;

        execList.Clear;
        pragma Debug( Dbg( "Processes.Managers.Ticker_Task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            pragma Debug( Dbg( "Processes.Managers.Ticker_Task Exception: ...", D_PROCS, ERROR ) );
            pragma Debug( Dbg( e, D_PROCS, ERROR ) );
            A_Game_Application(Get_Application).Stop( ERROR_UNEXPECTED_EXCEPTION );
    end Ticker_Task;

    ----------------------------------------------------------------------------

    procedure Delete( ticker : in out A_Ticker_Task ) is
        procedure Free is new Ada.Unchecked_Deallocation( Ticker_Task, A_Ticker_Task );
    begin
        Free( ticker );
    end Delete;

end Processes.Managers;
