
with Ada.Strings;                       use Ada.Strings;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Long_Float_Text_IO;
with Ada.Long_Long_Float_Text_IO;

package body Support.Real_Time is

    function Elapsed( t : Timer_Type ) return Time_Span is
    begin
        if t.running then
            -- timer is currently running
            return Clock - t.started;
        elsif t.started > Time_First then
            -- timer is paused, return elapsed before pause
            return t.paused - t.started;
        else
            -- timer hasn't been started
            return Time_Span_Zero;
        end if;
    end Elapsed;

    ----------------------------------------------------------------------------

    function Is_Running( t : Timer_Type ) return Boolean is
    begin
        return t.running;
    end Is_Running;

    ----------------------------------------------------------------------------

    procedure Pause( t : in out Timer_Type ) is
    begin
        if t.running then
            t.paused := Clock;
            t.running := False;
        end if;
    end Pause;

    ----------------------------------------------------------------------------

    procedure Restart( t : in out Timer_Type ) is
    begin
        t.running := False;         -- stop it
        t.started := Time_First;    -- clear it
        Start( t );                 -- start it again
    end Restart;

    ----------------------------------------------------------------------------

    procedure Start( t : in out Timer_Type ) is
    begin
        if not t.running then
            if t.started = Time_First then
                -- start timer for the first time
                t.started := Clock;
            else
                -- timer is paused, resume it
                t.started := Clock - (t.paused - t.started);
            end if;
            t.running := True;
        end if;
    end Start;

    --==========================================================================

    function Is_Updated( t : Frame_Timer ) return Boolean is
    begin
        return t.updated;
    end Is_Updated;

    ----------------------------------------------------------------------------

    function Rate( t : access Frame_Timer ) return Natural is
    begin
        t.updated := False;
        return t.fps;
    end Rate;

    ----------------------------------------------------------------------------

    procedure Rate( t : in out Frame_Timer; fps : out Natural ) is
    begin
        t.updated := False;
        fps := t.fps;
    end Rate;

    ----------------------------------------------------------------------------

    procedure Set_Update_Delay( t : in out Frame_Timer; ts : Time_Span ) is
    begin
        t.updateDelay := ts;
    end Set_Update_Delay;

    ----------------------------------------------------------------------------

    procedure Tick( t : in out Frame_Timer ) is
        now : constant Time := Clock;
    begin
        t.ticks := t.ticks + 1;
        if now - t.updateDelay >= t.lastUpdate then
            t.fps := Seconds( 1 ) / ((now - t.lastUpdate) / t.ticks);
            t.ticks := 0;
            t.lastUpdate := now;
            t.updated := True;
        end if;
    end Tick;

    --==========================================================================

    function Format( ts : Time_Span ) return String is
        secs : constant Integer := To_Seconds( ts );
    begin
        if secs < 1 then
            return Image( To_Milliseconds( ts ) ) & "[ms]";
        elsif secs < 60 then
            return Image( secs ) & "[s]";
        else
            return Image( To_Minutes( ts ) ) & "[min]";
        end if;
    end Format;

    ----------------------------------------------------------------------------

    function To_Float( ts : Time_Span ) return Long_Float is
    begin
        return Long_Float(To_Duration( ts ));
    end To_Float;

    ----------------------------------------------------------------------------

    function To_Microseconds( ts : Time_Span ) return Natural is
    begin
        return Natural(To_Duration( ts ) * 1_000_000.0);
    end To_Microseconds;

    ----------------------------------------------------------------------------

    function To_Milliseconds( ts : Time_Span ) return Natural is
    begin
        return Natural(To_Duration( ts ) * 1_000.0);
    end To_Milliseconds;

    ----------------------------------------------------------------------------

    function To_Minutes( ts : Time_Span ) return Natural is
    begin
        return Natural(To_Duration( ts ) / 60.0);
    end To_Minutes;

    ----------------------------------------------------------------------------

    function To_Seconds( ts : Time_Span ) return Natural is
    begin
        return Natural(To_Duration( ts ));
    end To_Seconds;

    ----------------------------------------------------------------------------

    function To_String( t : Time ) return String is
        secs : Seconds_Count;
        span : Time_Span;
        num  : Long_Long_Float;
        txt  : String(1..18) := (others => ' ');
    begin
        Split( t, secs, span );
        num := Long_Long_Float(secs) + Long_Long_Float(To_Duration( span ));
        Ada.Long_Long_Float_Text_IO.Put( txt, num, Aft => 3, Exp => 0 );
        return Ada.Strings.Fixed.Trim( txt, Both );
    end To_String;

    ----------------------------------------------------------------------------

    function To_String( ts : Time_Span; precision : Natural := 3 ) return String is
        num : Long_Float;
        txt : String(1..18) := (others => ' ');
    begin
        num := Long_Float(To_Duration( ts ));
        Ada.Long_Float_Text_IO.Put( txt, num, Aft => precision, Exp => 0 );
        return Ada.Strings.Fixed.Trim( txt, Both );
    end To_String;

    --==========================================================================

    function Time_Input( stream : access Root_Stream_Type'Class ) return Time is
    begin
        if Boolean'Input( stream ) then
            -- time is in the future
            return Clock + Time_Span'Input( stream );
        else
            -- time is in the past
            return Clock - Time_Span'Input( stream );
        end if;
    end Time_Input;

    ----------------------------------------------------------------------------

    procedure Time_Output( stream : access Root_Stream_Type'Class; t : Time ) is
        now    : constant Time := Clock;
        future : constant Boolean := t > now;
    begin
        Boolean'Output( stream, future );
        if future then
            Time_Span'Output( stream, t - now );
        else
            Time_Span'Output( stream, now - t );
        end if;
    end Time_Output;

end Support.Real_Time;
