
with Ada.Containers.Doubly_Linked_Lists;
with Ada.Task_Identification;           use Ada.Task_Identification;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Support.Real_Time;                 use Support.Real_Time;
with Tiles.Indices;                     use Tiles.Indices;

package body Tiles.Libraries.Loading is

    initialized : Boolean := False;
    finalized   : Boolean := False;

    --==========================================================================

    type Queued_Library is
        record
            lib     : A_Tile_Library := null;
            archive : A_Archive := null;
        end record;

    package Queued_Libraries is new Ada.Containers.Doubly_Linked_Lists( Queued_Library, "=" );
    use Queued_Libraries;

    ----------------------------------------------------------------------------

    protected Library_Queue is
        procedure Add( lib : A_Tile_Library; archive : in out A_Archive; success : out Boolean );
        entry Remove( lib : out A_Tile_Library; archive : out A_Archive );
        procedure Stop;
    private
        queue   : Queued_Libraries.List;
        ready   : Boolean := False;
        stopped : Boolean := False;
    end Library_Queue;

    protected body Library_Queue is

        procedure Add( lib : A_Tile_Library; archive : in out A_Archive; success : out Boolean ) is
        begin
            if not stopped then
                queue.Append( Queued_Library'(lib, archive) );
                archive := null;
                success := True;
                ready := True;
            else
                Delete( archive );
                success := False;
            end if;
        end Add;

        ------------------------------------------------------------------------

        entry Remove( lib : out A_Tile_Library; archive : out A_Archive ) when ready is
            qlib : Queued_Library;
        begin
            if not stopped then
                qlib := queue.First_Element;
                queue.Delete_First;
                lib := qlib.lib;
                archive := qlib.archive;
                ready := not queue.Is_Empty;
            else
                lib := null;
                archive := null;
            end if;
        end Remove;

        ------------------------------------------------------------------------

        procedure Stop is

            procedure Clear_Library( pos : Queued_Libraries.Cursor ) is
                qlib : Queued_Library := Queued_Libraries.Element( pos );
            begin
                Unload_Library( qlib.lib );
                Delete( qlib.archive );
            end Clear_Library;

        begin
            queue.Iterate( Clear_Library'Access );
            queue.Clear;
            ready := True;
            stopped := True;
        end Stop;

    end Library_Queue;

    --==========================================================================

    task type Loading_Task;
    type A_Loading_Task is access all Loading_Task;

    loader : A_Loading_Task := null;

    task body Loading_Task is
        lib     : A_Tile_Library;
        archive : A_Archive;
        timer   : Timer_Type;
    begin
        pragma Debug( Dbg( "Loading_Task starting", D_TILES, INFO ) );

        loop
            Library_Queue.Remove( lib, archive );
            exit when lib = null or else archive = null;

            Restart( timer );
            Load_Images( lib.index, archive );
            pragma Debug( Dbg( "Library <" & lib.Get_Name & "> load time: " &
                               Image( To_Milliseconds( Elapsed( timer ) ) ) & "[ms]",
                               D_TILES, INFO ) );

            Load_Complete( lib.index );
            lib.loadState.Complete;
            Unload_Library( lib );
            Delete( archive );
            Resources.Loading_Ends;          -- completed successful
        end loop;

        pragma Debug( Dbg( "Tiles.Libraries.Loading.Loading_Task complete", D_INIT, INFO ) );
    exception
        when e : others =>
            pragma Debug( Dbg( "Loading_Task Exception: ...", D_TILES, ERROR ) );
            pragma Debug( Dbg( e, D_TILES, ERROR ) );
            if lib /= null then
                -- free up threads waiting for the library to load even though
                -- we didn't finish loading properly.
                Load_Complete( lib.index );
                lib.loadState.Complete;
                Unload_Library( lib );
                Resources.Loading_Ends;      -- loading failed
            end if;
            Library_Queue.Stop;
    end Loading_Task;

    ----------------------------------------------------------------------------

    procedure Delete is new Ada.Unchecked_Deallocation( Loading_Task, A_Loading_Task );

    --==========================================================================

    procedure Initialize is
    begin
        if not initialized then
            loader := new Loading_Task;
            initialized := True;
        end if;
    end Initialize;

    ----------------------------------------------------------------------------

    procedure Finalize is
    begin
        if initialized and then not finalized then
            Library_Queue.Stop;

            -- Ten second timeout (we might be loading a large library)
            for count in 1..100 loop
                exit when Is_Terminated( loader.all'Identity );
                delay 0.1;
            end loop;
            if Is_Terminated( loader.all'Identity ) then
                Delete( loader );
            else
                pragma Debug( Dbg( "Tiles.Libraries.Loading.Finalize: " &
                                   "Loading_Task failed to stop", D_INIT, ERROR ) );
                null;
            end if;

            finalized := True;
        end if;
    end Finalize;

    ----------------------------------------------------------------------------

    function Load_From_Disk( name : String ) return A_Tile_Library is
        filename  : constant String := Add_Extension( name, Library_Extension, force => True );
        indexFile : constant String := Add_Extension( name, Index_Extension, force => True );
        lib       : A_Tile_Library := null;
        resource  : A_Resource_File := null;
        archive   : A_Archive := null;
        success   : Boolean := False;
    begin
        if name'Length = 0 then
            return null;
        end if;

        Resources.Loading_Begins;
        lib := Create_Tile_Library;
        lib.name := To_Unbounded_String( name );
        lib.dir := To_Unbounded_String( "" );
        resource := Load_Resource( filename, "graphics" );
        if resource /= null then
            archive := Load_Archive( resource );  -- may raise exception
            Unload_Resource( resource );
            lib.index := Load_Index( archive, indexFile );
            success := lib.index /= null;
            if success then
                -- hand the library off to the loader task
                Library_Queue.Add( lib, archive, success );
                pragma Debug( not success,
                              Dbg( "Error: Unable to load library <" & filename & ">: " &
                                   "Loader stopped", D_TILES, ERROR ) );
            else
                null;
                pragma Debug( Dbg( "Error: Index <" & indexFile &
                                   "> not found in library <" & filename & ">",
                                   D_TILES, ERROR ) );
            end if;
        end if;

        if not success then
            Delete( archive );
            Delete( A_Limited_Object(lib) );
            Resources.Loading_Ends;       -- loading failed
        end if;

        return lib;
    exception
        when e : others =>
            -- unable to open archive
            pragma Debug( Dbg( "Exception: Unable to load library <" & filename & ">",
                               D_TILES, ERROR ) );
            pragma Debug( Dbg( e, D_TILES, ERROR ) );
            Delete( archive );
            Unload_Resource( resource );
            Delete( A_Limited_Object(lib) );
            Resources.Loading_Ends;       -- loading failed
            return null;
    end Load_From_Disk;

end Tiles.Libraries.Loading;
