
with Ada.Unchecked_Deallocation;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Game_Views;                        use Game_Views;
with Widgets.Containers;                use Widgets.Containers;

package body Widgets.Containers.Scenes is

    use Widget_Lists;

    ----------------------------------------------------------------------------

    overriding
    procedure Add( this    : access Scene;
                   child   : in out A_Widget;
                   consume : Boolean := True ) is
    begin
        if child.all in Sprite'Class then
            this.spriteMap.Insert( A_Sprite(child).Get_Entity_Id, A_Sprite(child) );
        end if;
        Container(this.all).Add( child, consume );
    end Add;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Scene;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Container(this.all).Construct( view, id );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Frame_Changed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Attribute_Changed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Created" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Deleted" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Entity_Resized" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "New_World" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Tile_Changed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_View.Get_Corral.Add_Listener( A_Event_Listener(this), "Zoom" );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Clear_Selection( this : access Scene ) is
        crsr : Widget_Sets.Cursor := this.selectedSet.First;
    begin
        while Has_Element( crsr ) loop
            A_Sprite(Widget_Sets.Element( crsr )).Set_Selected( False );
            crsr := Next( crsr );
        end loop;
        this.selectedSet.Clear;
    end Clear_Selection;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Scene ) is
    begin
        this.Clear_Selection;
        Delete( this.lightLevels );
        Delete( this.visibleLayers );
        Delete( this.bitmapCache );
        Delete( this.mapLayers );
        Unload_Library( this.tileLib );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Zoom" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "World_Property_Changed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Tile_Changed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "New_World" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Resized" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Moved" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Deleted" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Created" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Entity_Attribute_Changed" );
        this.Get_View.Get_Corral.Remove_Listener( A_Event_Listener(this), "Frame_Changed" );
        Container(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete_Child( this : access Scene; child : in out A_Widget ) is
    begin
        if child.all in Sprite'Class then
            A_Scene(this).Unselect( child );
            this.spriteMap.Delete( A_Sprite(child).Get_Entity_Id );
        end if;
        Container(this.all).Delete_Child( child );
    end Delete_Child;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete_Children( this : access Scene ) is
    begin
        this.Clear_Selection;
        this.spriteMap.Clear;
        Container(this.all).Delete_Children;
    end Delete_Children;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Scene; dc : Drawing_Context ) is
        vwidth  : constant Integer := this.vx2 - this.vx1 + 1;
        vheight : constant Integer := this.vy2 - this.vy1 + 1;

        -- tile range of the visible scene (tile coords)
        tileX1, tileY1,
        tileX2, tileY2 : Integer;

        -- location in world coordinates to start drawing a row
        startX, startY : Integer;
    begin
        if Get_Layer( dc ) = Background then
            Clear_To_Color( dc, this.colors(Background) );
        end if;

        if not this.hasData or else this.tileLib = null then
            return;
        elsif this.waitForLoad and then not this.tileLib.Is_Loaded then
            -- ensure that this widget will be drawn again right away
            A_Widget(this).Set_Dirty;
            return;
        end if;

        tileX1 := Integer(this.vx1 / this.tileWidth);
        tileY1 := Integer(this.vy1 / this.tileWidth);
        tileX2 := Integer'Min( tileX1 + vwidth / this.tileWidth + 1, this.mapWidth - 1 );
        tileY2 := Integer'Min( tileY1 + vheight / this.tileWidth + 1, this.mapHeight - 1 );
        startX := tileX1 * this.tileWidth;
        startY := tileY1 * this.tileWidth;

        if Get_Layer( dc ) = Background then
            for l in this.mapLayers'First..Integer'Min( this.spriteLayer, this.mapLayers'Last ) loop
                if this.visibleLayers(l) then
                    A_Scene(this).Draw_Layer( dc, l, startX, startY,
                                              tileX1, tileY1, tileX2, tileY2 );
               end if;
            end loop;
        end if;

        if Get_Layer( dc ) = Foreground or else not this.drawForeground then
            for l in this.spriteLayer+1..this.mapLayers'Last loop
                if this.visibleLayers(l) then
                    A_Scene(this).Draw_Layer( dc, l, startX, startY,
                                              tileX1, tileY1, tileX2, tileY2 );
               end if;
            end loop;
            A_Scene(this).Draw_Layer_Overlay( dc, startX, startY,
                                              tileX1, tileY1, tileX2, tileY2 );
        end if;
    end Draw_Content;

    ----------------------------------------------------------------------------

    not overriding
    procedure Draw_Layer_Overlay( this   : access Scene;
                                  dc     : Drawing_Context;
                                  startX,
                                  startY : Integer;
                                  tileX1,
                                  tileY1,
                                  tileX2,
                                  tileY2 : Integer ) is
    begin
        null;
    end Draw_Layer_Overlay;

    ----------------------------------------------------------------------------

    function Find_Sprite( this : not null access Scene'Class;
                          id   : Entity_Id ) return A_Sprite is
        pos : constant Sprite_Maps.Cursor := this.spriteMap.Find( id );
    begin
        if Has_Element( pos ) then
            return Element( pos );
        end if;
        return null;
    end Find_Sprite;

    ----------------------------------------------------------------------------

    procedure Focus_On( this     : not null access Scene'Class;
                        x, y     : Float;
                        centered : Boolean := False ) is
        newX   : Integer := this.vx1;
        newY   : Integer := this.vy1;
        slackX : Float := this.wSlackX;
        slackY : Float := this.wSlackY;
    begin
        if centered then
            slackX := 0.0;
            slackY := 0.0;
        end if;

        if this.wFocusX > x + slackX then
            this.wFocusX := x + slackX;       -- move focus left
            newX := Integer'Max( Integer(this.wFocusX - Float(this.Get_Width / 2) / this.zoom), 0 );
        elsif this.wFocusX < x - slackX then
            this.wFocusX := x - slackX;       -- move focus right
            newX := Integer'Max( Integer(this.wFocusX - Float(this.Get_Width / 2) / this.zoom), 0 );
        end if;
        if this.wFocusY > y + slackY then
            this.wFocusY := y + slackY;       -- move focus up
            newY := Integer'Max( Integer(this.wFocusY - Float(this.Get_Height / 2) / this.zoom), 0 );
        elsif this.wFocusY < y - slackY then
            this.wFocusY := y - slackY;       -- move focus down
            newY := Integer'Max( Integer(this.wFocusY - Float(this.Get_Height / 2) / this.zoom), 0 );
        end if;

        this.Scroll_To( newX, newY );
    end Focus_On;

    ----------------------------------------------------------------------------

    function Get_Height_Tiles( this : not null access Scene'Class ) return Natural is
    begin
        return this.mapHeight;
    end Get_Height_Tiles;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Height( this : access Scene ) return Natural is
    begin
        if this.minHeight > 0 then
            return this.minHeight;
        elsif this.hasData then
            return this.tileWidth * this.mapHeight;
        else
            return 0;
        end if;
    end Get_Min_Height;

    ----------------------------------------------------------------------------

    overriding
    function Get_Min_Width( this : access Scene ) return Natural is
    begin
        if this.minWidth > 0 then
            return this.minWidth;
        elsif this.hasData then
            return this.tileWidth * this.mapWidth;
        else
            return 0;
        end if;
    end Get_Min_Width;

    ----------------------------------------------------------------------------

    function Get_Width_Tiles( this : not null access Scene'Class ) return Natural is
    begin
        return this.mapWidth;
    end Get_Width_Tiles;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Attribute_Changed( this : access Scene;
                                               evt  : not null A_Entity_Attribute_Changed_Event ) is
        sprite : A_Sprite;
        val    : A_Value;
    begin
        sprite := this.Find_Sprite( evt.Get_Id );
        if sprite /= null then
            evt.Copy_Value( val );
            sprite.Set_Attribute( evt.Get_Attribute, val );
        end if;
    end Handle_Entity_Attribute_Changed;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Created( this : access Scene;
                                     evt  : not null A_Entity_Created_Event ) is
        spr : A_Sprite;
    begin
        if not evt.Is_Metaphysical then
            spr := Create_Sprite( this.Get_View, evt.Get_Id, evt.Get_Class,
                                  evt.Get_X, evt.Get_Y, evt.Is_Physical,
                                  evt.Get_Lib_Name, evt.Get_Frame );
            spr.Set_Visible( True );
            this.Add( A_Widget(spr) );
            if evt.Get_Id = this.target then
                this.Focus_On( evt.Get_X, evt.Get_Y, centered => True );
            end if;
        end if;
    end Handle_Entity_Created;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Deleted( this : access Scene;
                                     evt  : not null A_Entity_Deleted_Event ) is
        sprite : A_Sprite := this.Find_Sprite( evt.Get_Id );
    begin
        if sprite /= null then
            A_Scene(this).Delete_Child( A_Widget(sprite) );
        end if;
    end Handle_Entity_Deleted;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Moved( this : access Scene;
                                   evt  : not null A_Entity_Moved_Event ) is
        sprite : A_Sprite;
    begin
        if evt.Get_Id = this.target then
            -- update the scene's position
            this.Focus_On( evt.Get_X, evt.Get_Y );
        end if;

        sprite := this.Find_Sprite( evt.Get_Id );
        if sprite /= null and then sprite.Is_Updatable then
            sprite.Center_At( evt.Get_X, evt.Get_Y );
        end if;
    end Handle_Entity_Moved;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Entity_Resized( this : access Scene;
                                     evt  : not null A_Entity_Resized_Event ) is
        sprite : constant A_Sprite := this.Find_Sprite( evt.Get_Id );
    begin
        if sprite /= null and then sprite.Is_Updatable then
            sprite.Resize( evt.Get_Width, evt.Get_Height );
        end if;
    end Handle_Entity_Resized;

    ----------------------------------------------------------------------------

    -- if evt is returned as null then the event was consumed
    overriding
    procedure Handle_Event( this : access Scene;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
        pragma Unreferenced( resp );
    begin
        if evt.Get_Id = FRAME_CHANGED_ID then
            A_Scene(this).Handle_Frame_Changed( A_Frame_Changed_Event(evt) );
        elsif evt.Get_Id = ENTITY_MOVED_ID then
            A_Scene(this).Handle_Entity_Moved( A_Entity_Moved_Event(evt) );
        elsif evt.Get_Id = TILE_CHANGED_ID then
            A_Scene(this).Handle_Tile_Changed( A_Tile_Changed_Event(evt) );
        elsif evt.Get_Id = NEW_WORLD_ID then
            A_Scene(this).Handle_New_World( A_New_World_Event(evt) );
        elsif evt.Get_Id = WORLD_PROPERTY_CHANGED_ID then
            A_Scene(this).Handle_World_Property_Changed( A_World_Property_Changed_Event(evt) );
        elsif evt.Get_Id = ENTITY_ATTRIBUTE_CHANGED_ID then
            A_Scene(this).Handle_Entity_Attribute_Changed( A_Entity_Attribute_Changed_Event(evt) );
        elsif evt.Get_Id = ENTITY_CREATED_ID then
            A_Scene(this).Handle_Entity_Created( A_Entity_Created_Event(evt) );
        elsif evt.Get_Id = ENTITY_DELETED_ID then
            A_Scene(this).Handle_Entity_Deleted( A_Entity_Deleted_Event(evt) );
        elsif evt.Get_Id = ENTITY_RESIZED_ID then
            A_Scene(this).Handle_Entity_Resized( A_Entity_Resized_Event(evt) );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Frame_Changed( this : access Scene;
                                    evt  : not null A_Frame_Changed_Event ) is
        pos   : Widget_Lists.Cursor := this.children.First;
        child : A_Widget;
    begin
        while Has_Element( pos ) loop
            child := Element( pos );
            if child.all in Sprite'Class and then
               A_Sprite(child).Get_Entity_Id = evt.Get_Id and then
               A_Sprite(child).Is_Updatable
            then
                A_Sprite(child).Set_Frame( evt.Get_Frame );
                exit;
            end if;
            pos := Next( pos );
        end loop;
    end Handle_Frame_Changed;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_New_World( this : access Scene;
                                evt  : not null A_New_World_Event ) is
    begin
        this.Delete_Children;
        Delete( this.lightLevels );
        Delete( this.visibleLayers );
        Delete( this.bitmapCache );
        Delete( this.mapLayers );
        Unload_Library( this.tileLib );

        this.mapWidth := evt.Get_Width;
        this.mapHeight := evt.Get_Height;
        this.tileWidth := evt.Get_Tile_Width;
        this.mapLayers := Copy( evt.Get_Layers );
        this.bitmapCache := new Bitmap_Cache(this.mapLayers'Range);
        for i in this.bitmapCache'Range loop
            this.bitmapCache(i) := new Cached_Bitmap_Array(this.mapLayers(i).data'Range);
        end loop;
        this.visibleLayers := new Boolean_Array(this.mapLayers'Range);
        for i in this.visibleLayers'Range loop
            this.visibleLayers(i) := True;
        end loop;
        this.lightLevels := new Byte_Array(this.mapLayers'Range);
        for i in this.lightLevels'Range loop
            this.lightLevels(i) := 255;
        end loop;
        this.spriteLayer := this.mapLayers'Last;

        this.tileLib := Load_Library( evt.Get_Library_Name );
        this.hasData := True;

        A_Scene(this).Pack;
        this.Set_Zoom( 1.0 );
        this.Focus_On( 0.0, 0.0 );
    end Handle_New_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Resize( this : access Scene ) is
        wCenterX,
        wCenterY : Integer;
    begin
        Container(this.all).Handle_Resize;

        if this.target = INVALID_ID then
            -- Put the focus back in the center of the scene
            A_Widget(this).Translate_To_Content( this.x1 + this.Get_Width / 2,
                                                 this.y1 + this.Get_Height / 2,
                                                 wCenterX, wCenterY );
            A_Scene(this).Focus_On( Float(wCenterX), Float(wCenterY) );
        end if;
    end Handle_Resize;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Tile_Changed( this : access Scene;
                                   evt  : not null A_Tile_Changed_Event ) is
    begin
        if this.hasData then
            this.mapLayers(evt.Get_Layer).data(
                this.mapLayers(evt.Get_Layer).data'First +
                evt.Get_Y * this.mapWidth + evt.Get_X) := evt.Get_Tile_ID;
            if this.Is_Tile_Visible( evt.Get_Layer, evt.Get_X, evt.Get_Y ) then
                A_Widget(this).Set_Dirty;
            end if;
        end if;
    end Handle_Tile_Changed;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_World_Property_Changed( this : access Scene;
                                             evt  : not null A_World_Property_Changed_Event ) is
        pragma Unreferenced( this, evt );
    begin
        null;
    end Handle_World_Property_Changed;

    ----------------------------------------------------------------------------

    function Is_Tile_Visible( this  : not null access Scene'Class;
                              layer : Positive;
                              x, y  : Integer ) return Boolean is
        vwidth, vheight : Integer;
        tileX1, tileY1,
        tileX2, tileY2  : Integer;
    begin
        if this.visibleLayers(layer) then
            vwidth := this.vx2 - this.vx1 + 1;
            vheight := this.vy2 - this.vy1 + 1;
            tileX1 := Integer(this.vx1 / this.tileWidth);
            tileY1 := Integer(this.vy1 / this.tileWidth);
            tileX2 := Integer'Min( tileX1 + vwidth / this.tileWidth + 1, this.mapWidth - 1 );
            tileY2 := Integer'Min( tileY1 + vheight / this.tileWidth + 1, this.mapHeight - 1 );
            return x >= tileX1 and then x <= tileX2 and then
                   y >= tileY1 and then y <= tileY2;
        end if;
        return False;
    end Is_Tile_Visible;

    ----------------------------------------------------------------------------

    overriding
    procedure Remove( this : access Scene; child : not null A_Widget ) is
    begin
        if child.all in Sprite'Class then
            A_Scene(this).Unselect( child );
            this.spriteMap.Delete( A_Sprite(child).Get_Entity_Id );
        end if;
        Container(this.all).Remove( child );
    end Remove;

    ----------------------------------------------------------------------------

    procedure Scroll( this : not null access Scene'Class; x, y : Float ) is
        tx, ty : Float;     -- scene's target's location
        fx, fy : Float;     -- new focus location
        sprite : A_Sprite;  -- scene's target sprite
    begin
        sprite := this.Find_Sprite( this.target );
        if sprite /= null then
            tx := sprite.Get_X;
            ty := sprite.Get_Y;
        else
            tx := this.wFocusX;
            ty := this.wFocusY;
        end if;

        if x > 0.0 then
            fx := Float'Min( this.wFocusX + x, tx + this.wSlackX );
        else
            fx := Float'Max( this.wFocusX + x, tx - this.wSlackX );
        end if;
        if y > 0.0 then
            fy := Float'Min( this.wFocusY + y, ty + this.wSlackY );
        else
            fy := Float'Max( this.wFocusY + y, ty - this.wSlackY );
        end if;

        this.Focus_On( fx, fy, centered => True );
    end Scroll;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Selection( this : access Scene; child : A_Widget ) is
    begin
        this.Clear_Selection;
        if child /= null then
            this.selectedSet.Insert( child );
            A_Sprite(child).Set_Selected( True );
        end if;
    end Set_Selection;

    ----------------------------------------------------------------------------

    not overriding
    procedure Set_Target( this : access Scene; target : Entity_Id ) is
        sprite : A_Sprite;
    begin
        this.target := target;
        pragma Debug( Dbg( "Scene: Target =" & Entity_Id'Image( this.target ), D_GUI, INFO ) );
        sprite := this.Find_Sprite( target );
        if sprite /= null then
            this.Focus_On( sprite.Get_X, sprite.Get_Y, centered => True );
        end if;
    end Set_Target;

    ----------------------------------------------------------------------------

    not overriding
    procedure Unselect( this : access Scene; child : not null A_Widget ) is
        crsr : Widget_Sets.Cursor := this.selectedSet.Find( child );
    begin
        if Has_Element( crsr ) then
            this.selectedSet.Delete( crsr );
            A_Sprite(child).Set_Selected( False );
        end if;
    end Unselect;

    --==========================================================================

    procedure Delete( ba : in out A_Boolean_Array ) is
        procedure Free is new Ada.Unchecked_Deallocation( Boolean_Array, A_Boolean_Array );
    begin
        Free( ba );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Delete( ba : in out A_Byte_Array ) is
        procedure Free is new Ada.Unchecked_Deallocation( Byte_Array, A_Byte_Array );
    begin
        Free( ba );
    end Delete;

    ----------------------------------------------------------------------------

    procedure Delete( bc : in out A_Bitmap_Cache ) is
        procedure Free is new Ada.Unchecked_Deallocation( Cached_Bitmap_Array, A_Cached_Bitmap_Array );
        procedure Free is new Ada.Unchecked_Deallocation( Bitmap_Cache, A_Bitmap_Cache );
    begin
        if bc /= null then
            for i in bc'Range loop
                Free( bc(i) );
            end loop;
        end if;
        Free( bc );
    end Delete;

end Widgets.Containers.Scenes;
