
with Ada.Unchecked_Deallocation;
--with Debugging;                         use Debugging;
with Font_API;                          use Font_API;
with Mouse;                             use Mouse;
with Support;                           use Support;
with Widgets.Containers.Pulldown_Menus; use Widgets.Containers.Pulldown_Menus;

package body Widgets.Menubars is

    -- the height of the menubar is proportional to font height
    HEIGHT_MULTIPLIER : constant Float := 1.5;

    ----------------------------------------------------------------------------

    function Create_Menubar( view : not null access Game_Views.Game_View'Class;
                             id   : String ) return A_Menubar is
        this : A_Menubar := new Menubar;
    begin
        this.Construct( view, id );
        return this;
    exception
        when others =>
            Delete( A_Widget(this) );
            raise;
    end Create_Menubar;

    ----------------------------------------------------------------------------

    not overriding
    procedure Add( this : access Menubar; pulldown : in out A_Widget ) is
        PADCHAR : constant Character := 'O';
        newMenu : A_Menu;
    begin
        if pulldown /= null and then pulldown.all in Pulldown_Menu'Class then
            newMenu := new Menu;
            newMenu.pulldown := pulldown;
            if not this.menus.Is_Empty then
                newMenu.x := this.menus.Last_Element.x +
                             this.menus.Last_Element.width;
            end if;
            newMenu.text := To_Unbounded_String( A_Pulldown_Menu(newMenu.pulldown).Get_Text );
            -- the two O's added onto the text length calculation are for
            -- horizontal padding. the width of the 'O' character is the
            -- padding before and after the menu text.
            newMenu.width := Themes.Get.Get_Text_Width( PADCHAR & To_String( newMenu.text ) & PADCHAR, To_String( this.fontName ), this.fontSize );
            newMenu.pulldown.Add_Listener( A_Visibility_Listener(this) );
            this.menus.Append( newMenu );
            A_Widget(this).Set_Dirty;
        elsif pulldown = null then
            -- can't add null
            raise Constraint_Error with "Widget is null";
        else
            -- can't add widgets of the wrong type
            raise Constraint_Error with "Invalid widget type";
        end if;

        pulldown := null;
    end Add;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Menubar;
                         view : not null access Game_Views.Game_View'Class;
                         id   : String ) is
    begin
        Widget(this.all).Construct( view, id );
        this.zoomable := False;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Menubar ) is
        m    : A_Menu;
        crsr : Menu_Lists.Cursor := this.menus.First;
    begin
        while Has_Element( crsr ) loop
            m := Element( crsr );
            m.pulldown.Remove_Listener( A_Visibility_Listener(this) );
            Delete( m );
            crsr := Next( crsr );
        end loop;
        this.menus.Clear;
        Widget(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    overriding
    procedure Draw_Content( this : access Menubar; dc : Drawing_Context ) is
        theme     : constant A_Theme := Themes.Get;
        font      : constant Font_Type := theme.Get_Font( To_String( this.fontName ), this.fontSize );
        text_y    : constant Integer := this.Get_Height / 2 - Text_Height( font ) / 2;
        padding_x : constant Integer := Text_Length( font, "O" );

        ------------------------------------------------------------------------

        procedure Draw_Menu( pos : Menu_Lists.Cursor ) is
            m      : constant A_Menu := Element( pos );
            border : Border_Type := this.border;
        begin
            Textout( dc, font, To_String( m.text ), m.x + padding_x, text_y,
                     this.colors(Text), this.fontSmooth );

            if m.active then
                border := Border_Sunk;
            elsif m.hover then
                border := Border_Raised;
            end if;

            if border /= Border_None then
                theme.Draw_Border( dc, border,
                                   m.x, 0,
                                   m.x + m.width - 1, this.cheight - 2,
                                   this.colors(Foreground) );
            end if;
        end Draw_Menu;

        ------------------------------------------------------------------------

        bw : constant Natural := theme.Border_Width;
    begin
        Rectfill( dc, 0, 0, this.cwidth - 1, this.cheight - 2,
                  this.colors(Foreground) );

        -- draw a border such that only the bottom is actually visible along the
        -- bottom of the menubar
        theme.Draw_Border( dc, Border_Raised,
                           -bw, -bw, this.cwidth - 1 + bw, this.cheight - 1,
                           this.colors(Foreground) );

        this.menus.Iterate( Draw_Menu'Access );
    end Draw_Content;

    ----------------------------------------------------------------------------

    procedure Handle_Action( this   : access Menubar;
                             action : A_Visibility_Action ) is
        pulldown : constant A_Widget := A_Widget(action.Get_Source);
        crsr     : Menu_Lists.Cursor := this.menus.First;
        m        : A_Menu;
    begin
        if action.Get_Id = Change_Visibility and then not action.Is_Visible then
            while Has_Element( crsr ) loop
                m := Element( crsr );
                if m.pulldown = pulldown then
                    m.active := False;
                    A_Widget(this).Set_Dirty;
                    exit;
                end if;
                crsr := Next( crsr );
            end loop;
        end if;
    end Handle_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Exit( this : access Menubar ) is
        crsr : Menu_Lists.Cursor := this.menus.First;
    begin
        while Has_Element( crsr ) loop
            Element( crsr ).hover := False;
            crsr := Next( crsr );
        end loop;
        A_Widget(this).Set_Dirty;
    end Handle_Exit;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Move( this : access Menubar;
                                 evt  : not null A_Mouse_Event ) is
        ex   : constant Integer := evt.Get_X;
        ey   : constant Integer := evt.Get_Y;
        crsr : Menu_Lists.Cursor;
        m    : A_Menu;
    begin
        -- clear hovering states and deactivate popups
        crsr := this.menus.First;
        while Has_Element( crsr ) loop
            m := Element( crsr );
            -- if the mouse is not over this menu
            if not Contains( m.x, this.y1, m.x + m.width - 1, this.y2, ex, ey ) then
                -- if its hover is set, clear it
                if m.hover then
                    m.hover := False;
                    A_Widget(this).Set_Dirty;
                end if;
                -- if the mouse is being dragged and its not over this active
                -- menu, and the mouse is still within the menubar area
                -- vertically, disable it. the mouse vertical location is
                -- checked because the menu shouldn't be disabled if the mouse
                -- is being dragged below the menu bar (ie: onto the pulldown)
                if this.dragging and then m.active and then ey <= this.y2 then
                    A_Widget(this).Deactivate_Popup( m.pulldown );
                end if;
            end if;
            crsr := Next( crsr );
        end loop;

        -- set hovering states and activate popups
        crsr := this.menus.First;
        while Has_Element( crsr ) loop
            m := Element( crsr );
            if Contains( m.x, this.y1, m.x + m.width - 1, this.y2, ex, ey ) then
                if not m.hover then
                    m.hover := True;
                    A_Widget(this).Set_Dirty;
                    if this.dragging then
                        A_Widget(this).Activate_Popup( m.pulldown, Widget_Top_Left, m.x, this.y2, 0, 0, False );
                        m.active := True;
                    end if;
                end if;
                exit;
            end if;
            crsr := Next( crsr );
        end loop;
    end Handle_Mouse_Move;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Press( this : access Menubar;
                                  evt  : not null A_Mouse_Button_Event ) is
        ex   : constant Integer := evt.Get_X;
        ey   : constant Integer := evt.Get_Y;
        crsr : Menu_Lists.Cursor := this.menus.First;
        m    : A_Menu;
    begin
        if evt.Get_Button = Mouse_Left then
            while Has_Element( crsr ) loop
                m := Element( crsr );
                if Contains( m.x, this.y1, m.x + m.width - 1, this.y2, ex, ey ) then
                    -- activate or deactivate the popup
                    this.Activate_Popup( m.pulldown, Widget_Top_Left, m.x, this.y2, 0, 0, False );
                    m.active := True;
                    this.dragging := True;
                    exit;
                end if;
                crsr := Next( crsr );
            end loop;
        end if;
    end Handle_Mouse_Press;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Mouse_Release( this : access Menubar;
                                    evt  : not null A_Mouse_Button_Event ) is
    begin
        if evt.Get_Button = Mouse_Left then
            this.dragging := False;
        end if;
    end Handle_Mouse_Release;

    ----------------------------------------------------------------------------

    not overriding
    procedure Pulldown_Hidden( this     : access Menubar;
                               pulldown : not null A_Widget ) is
        crsr : Menu_Lists.Cursor := this.menus.First;
        m    : A_Menu;
    begin
        while Has_Element( crsr ) loop
            m := Element( crsr );
            if m.pulldown = pulldown then
                m.active := False;
                A_Widget(this).Set_Dirty;
                exit;
            end if;
            crsr := Next( crsr );
        end loop;
    end Pulldown_Hidden;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Parent( this : access Menubar; parent : A_Widget ) is
    begin
        this.Set_Layout_LTRB( 0, 0, 0, -Integer(Themes.Get.Get_Text_Height( To_String( this.fontName ), this.fontSize ) * HEIGHT_MULTIPLIER) );
        Widget(this.all).Set_Parent( parent );
    end Set_Parent;

    --==========================================================================

    procedure Delete( m : in out A_Menu ) is
        procedure Free is new Ada.Unchecked_Deallocation( Menu, A_Menu );
    begin
        if m /= null then
            Delete( A_Limited_Object(m.pulldown) );
            Free( m );
        end if;
    end Delete;

end Widgets.Menubars;
