
with Ada.Command_Line;                  use Ada.Command_Line;
with Ada.Exceptions;                    use Ada.Exceptions;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Games;                             use Games;
with Statuses;                          use Statuses;
with Tools.Pointers;                    use Tools.Pointers;
with Worlds.Importing;                  use Worlds.Importing;

pragma Elaborate_All( Games );

package body Games.Ked is

    function Create_Game return A_Game is
        this : constant A_Game := new Ked_Game;
    begin
        this.Construct;
        return this;
    end Create_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure Add_Event_Listeners( this : access Ked_Game ) is
    begin
        Game(this.all).Add_Event_Listeners;
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Apply_Tool_Map" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Create_World" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Import_World" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Impulse" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Resize_World" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Save_World" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_Tool" );
        this.Get_Corral.Add_Listener( A_Event_Listener(this), "Set_World_Property" );
    end Add_Event_Listeners;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Ked_Game ) is
    begin
        Game(this.all).Construct;
        this.tool := Create_Pointer;
    end Construct;

    ----------------------------------------------------------------------------

    overriding
    procedure Delete( this : in out Ked_Game ) is
    begin
        Delete( this.tool );
        Game(this).Delete;
    end Delete;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Apply_Tool( this : access Ked_Game;
                                 evt  : not null A_Apply_Tool_Map_Event ) is
    begin
        if not this.paused and then this.tool /= null then
            this.tool.Apply( evt.Get_Function, evt.Get_Modifiers, evt.Is_First,
                             this.world, evt.Get_X, evt.Get_Y, evt.Get_Layer );
        end if;
    end Handle_Apply_Tool;

    ----------------------------------------------------------------------------

    procedure Handle_Create_World( this : access Ked_Game;
                                   evt  : not null A_Create_World_Event;
                                   resp : out Response_Type ) is
        new_world : A_World;
    begin
        -- may raise an exception
        new_world := Create_World( evt.Get_Width,
                                   evt.Get_Height,
                                   evt.Get_Library_Name,
                                   evt.Get_Domain );
        this.Load_World( new_world );   -- consumes new_world
        Queue_World_Modified;
        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Create_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Handle_Event( this : access Ked_Game;
                            evt  : in out A_Event;
                            resp : out Response_Type ) is
    begin
        if evt.Get_Id = APPLY_TOOL_MAP_ID then
            this.Handle_Apply_Tool( A_Apply_Tool_Map_Event(evt) );
        elsif evt.Get_Id = IMPULSE_ID then
            this.Handle_Impulse( A_Impulse_Event(evt) );
        elsif evt.Get_Id = SET_TOOL_ID then
            this.Handle_Set_Tool( A_Set_Tool_Event(evt) );
        elsif evt.Get_Id = SAVE_WORLD_ID then
            this.Handle_Save_World( A_Save_World_Event(evt), resp );
        elsif evt.Get_Id = CREATE_WORLD_ID then
            this.Handle_Create_World( A_Create_World_Event(evt), resp );
        elsif evt.Get_Id = IMPORT_WORLD_ID then
            this.Handle_Import_World( A_Import_World_Event(evt), resp );
        elsif evt.Get_Id = RESIZE_WORLD_ID then
            this.Handle_Resize_World( A_Resize_World_Event(evt), resp );
        elsif evt.Get_Id = SET_WORLD_PROPERTY_ID then
            this.Handle_Set_World_Property( A_Set_World_Property_Event(evt) );
        else
            Game(this.all).Handle_Event( evt, resp );
        end if;
    end Handle_Event;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Import_World( this : access Ked_Game;
                                   evt  : not null A_Import_World_Event;
                                   resp : out Response_Type ) is
        imported : A_World;
    begin
        -- may raise an exception
        imported := Import_World( evt.Get_Filename,
                                      evt.Get_Library_Name,
                                      evt.Get_Domain,
                                      evt.Get_Tolerance );
        this.Load_World( imported );  -- consumes imported
        Queue_World_Modified;
        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Import_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Impulse( this : access Ked_Game;
                              evt  : not null A_Impulse_Event ) is
        crsr : Actor_Maps.Cursor;
    begin
        crsr := this.actorIndex.Find( evt.Get_Id );
        if Actor_Maps.Has_Element( crsr ) then
            Actor_Maps.Element( crsr ).Impulse( evt.Get_Impulse_Name );
        end if;
    end Handle_Impulse;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Resize_World( this : access Ked_Game;
                                   evt  : not null A_Resize_World_Event;
                                   resp : out Response_Type ) is
    begin
        if this.world /= null then
            this.world.Resize( evt.Get_Width, evt.Get_Height );
        else
            Set_Response( resp, ST_FAILED, "No world to resize" );
        end if;
        Set_Response( resp, ST_SUCCESS );
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Resize_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Save_World( this : access Ked_Game;
                                 evt  : not null A_Save_World_Event;
                                 resp : out Response_Type ) is
    begin
        if this.world /= null then
            this.world.Save( evt.Get_Filename );
            Set_Response( resp, ST_SUCCESS );
        else
            Set_Response( resp, ST_FAILED, "No world to save" );
        end if;
    exception
        when e : others =>
            Set_Response( resp, ST_FAILED, Exception_Message( e ) );
    end Handle_Save_World;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Set_Tool( this : access Ked_Game;
                               evt  : not null A_Set_Tool_Event ) is
    begin
        if not this.paused then
            this.tool := Copy( evt.Get_Tool );
        end if;
    end Handle_Set_Tool;

    ----------------------------------------------------------------------------

    not overriding
    procedure Handle_Set_World_Property( this : access Ked_Game;
                                         evt  : not null A_Set_World_Property_Event ) is
    begin
        if this.world /= null then
            this.world.Set_Property( evt.Get_Property_Name, evt.Get_Value );
        end if;
    end Handle_Set_World_Property;

    ----------------------------------------------------------------------------

    overriding
    procedure New_Game( this : access Ked_Game ) is
    begin
        Game(this.all).New_Game;

        -- load a filename passed on the command line or the default world
        for i in 1..Argument_Count loop
             declare
                 mapName : constant String := Argument( i );
             begin
                 if mapName(mapName'First) /= '-' then
                     this.Load_World( mapName );
                     return;
                 end if;
             exception
                 when e : others =>
                     null;
                     pragma Debug( Dbg( "New_Game: " &
                                        "Error loading world <" & mapName & ">: " &
                                        Exception_Message( e ), D_VIEW, ERROR ) );
             end;
        end loop;

        this.Load_World( "default" );
    exception
        when others =>
            null;
    end New_Game;

    ----------------------------------------------------------------------------

    overriding
    procedure Remove_Event_Listeners( this : access Ked_Game ) is
    begin
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_World_Property" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Set_Tool" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Save_World" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Resize_World" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Impulse" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Import_World" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Create_World" );
        this.Get_Corral.Remove_Listener( A_Event_Listener(this), "Apply_Tool_Map" );
        Game(this.all).Remove_Event_Listeners;
    end Remove_Event_Listeners;

begin

    Register_Allocator( Create_Game'Access );

end Games.Ked;
