
with Ada.Containers.Doubly_Linked_Lists;
with Ada.Exceptions;                    use Ada.Exceptions;
with Ada.Real_Time;                     use Ada.Real_Time;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Image_Files;               use Allegro.Image_Files;
with Allegro.Sprites;                   use Allegro.Sprites;
with Allegro.Truecolor;                 use Allegro.Truecolor;
with Debugging;                         use Debugging;
with Engine_Debugging;                  use Engine_Debugging;
with Maps;                              use Maps;
with Support;                           use Support;
with Themes.Palette;                    use Themes.Palette;
with Worlds;                            use Worlds;

package body Worlds.Importing is

    function Import_World( imagePath,
                           libName,
                           domain    : String;
                           tolerance : Natural ) return A_World is

        UPDATE_DELAY : constant Time_Span := Seconds( 3 );

        package Intlist is new Ada.Containers.Doubly_Linked_Lists(Integer, "=");
        use Intlist;

        ------------------------------------------------------------------------

        function Get_Tile_Width return Natural is
            map    : A_Map;
            result : Natural;
        begin
            map := Create_Map( 1, 1 );
            result := map.Tile_Width;
            Delete( map );
            return result;
        exception
            when others =>
                Delete( map );
                raise;
        end Get_Tile_Width;

        ------------------------------------------------------------------------

        function Is_Transparent( bmp : A_Bitmap ) return Boolean is
        begin
            if bmp = null then
                return False;
            end if;
            for y in 0..(Get_Height( bmp ) - 1) loop
                for x in 0..(Get_Width( bmp ) - 1) loop
                    if Getpixel( bmp, x, y ) = Themes.Palette.base_colors(Transparent) then
                        return True;
                    end if;
                end loop;
            end loop;
            return False;
        end Is_Transparent;

        ------------------------------------------------------------------------

        function Match( a, b : A_Bitmap ) return Boolean is
        begin
            if a = null and then b = null then
                return True;
            end if;

            if a = null or else b = null or else
               Get_Width( a ) /= Get_Width( b ) or else
               Get_Height( a ) /= Get_Height( b )
            then
                return False;
            end if;

            for y in 0..(Get_Height( a ) - 1) loop
                for x in 0..(Get_Width( a ) - 1) loop
                    if not Compare( Getpixel( a, x, y ),
                                    Getpixel( b, x, y ),
                                    tolerance )
                    then
                        return False;
                    end if;
                end loop;
            end loop;
            return True;
        end Match;
        ------------------------------------------------------------------------

        function Transparent_Match( bg, fg : A_Bitmap ) return Boolean is
            fgp : Color_Type;
        begin
            if bg = null or else fg = null or else
               Get_Width( bg ) /= Get_Width( fg ) or else
               Get_Height( bg ) /= Get_Height( fg )
            then
                return False;
            end if;

            for y in 0..(Get_Height( bg ) - 1) loop
                for x in 0..(Get_Width( bg ) - 1) loop
                    fgp := Getpixel( fg, x, y );
                    if fgp /= base_colors(Transparent) and then
                       not Compare( fgp, Getpixel( bg, x, y ), tolerance )
                    then
                        return False;
                    end if;
                end loop;
            end loop;
            return True;
        end Transparent_Match;

        ------------------------------------------------------------------------

        procedure Update_Progress( percent, secsRemaining : Integer ) is
        begin
            null;
            pragma Debug( Dbg( "  " &
                               Image( percent ) & "% : " &
                               Image( secsRemaining ) & "s remaining",
                               D_GAME, INFO ) );
        end Update_Progress;

        ------------------------------------------------------------------------

        bmp          : A_Bitmap := Load_Bitmap( imagePath, null );
        lib          : A_Tile_Library := Load_Library_Sync( libName );
        world        : A_World := null;
        tw           : Natural;
        start        : constant Time := Clock;
        lastUpdate   : Time := start;      -- time of last progress update
        progress     : Natural;
        lastProgress : Natural := 0;       -- progress at last update
        matched      : Natural := 0;       -- tiles successfully matched
        trans        : Intlist.List;
        tmpbmp       : A_Bitmap;
        clk          : Time;
        secsLeft     : Integer;
    begin
        if bmp = null then
            raise Constraint_Error with "Unable to read file <" & imagePath & ">";
        elsif lib = null then
            raise Constraint_Error with "Unable to load library <" & libName & ">";
        elsif not Valid_Domain( domain ) then
            raise Constraint_Error with "Invalid domain <" & domain & ">";
        end if;

        tw := Get_Tile_Width;
        if Get_Width( bmp ) mod tw /= 0 or else Get_Height( bmp ) mod tw /= 0 then
            raise Constraint_Error with "Bitmap dimensions must be a multiple of " &
                                        Image( tw );
        end if;

        tmpbmp := Create_Bitmap( tw, tw );
        for t in 1..lib.Get_Tile_Count loop
            if Is_Transparent( lib.Get_Bitmap_At_Slot( t ) ) then
                trans.Append( t );
            end if;
        end loop;

        pragma Debug( Dbg( "Matching tiles...", D_GAME, INFO ) );
        world := Create_World( Get_Width( bmp ) / tw, Get_Height( bmp ) / tw, libName, domain );
        for y in 0..(world.Get_Height_Tiles - 1) loop
            for x in 0..(world.Get_Width_Tiles - 1) loop

                declare
                    subbmp     : A_Bitmap := Create_Sub_Bitmap( bmp, x * tw, y * tw, tw, tw );
                    tilebmp    : A_Bitmap;
                    tpos       : Intlist.Cursor;
                    fgslot     : Natural := 0;
                    matched_fg,
                    matched_bg : Boolean := False;
                    found_fg,
                    found_bg   : Boolean := False;
                begin
                    -- try to match just a background
                    for bgtile in 1..lib.Get_Tile_Count loop
                        tilebmp := lib.Get_Bitmap_At_Slot( bgtile );
                        if tilebmp /= null then
                            if Match( subbmp, tilebmp ) then
                                world.Set_Tile( 1, x, y, lib.Get_Id_At_Slot( bgtile ), notify => False );
                                matched_bg := True;
                                found_bg := True;
                                exit;
                            end if;
                        end if;
                    end loop;

                    -- try to find a foreground tile that works
                    if not matched_bg then
                        tpos := trans.First;
                        while Has_Element( tpos ) loop
                            fgslot := Element( tpos );
                            if Transparent_Match( subbmp, lib.Get_Bitmap_At_Slot( fgslot ) ) then
                                -- try to combine a background with the foreground for a match
                                for bgslot in 1..lib.Get_Tile_Count loop
                                    tilebmp := lib.Get_Bitmap_At_Slot( bgslot );
                                    if tilebmp /= null then
                                        Blit( tilebmp, tmpbmp, 0, 0, 0, 0, tw, tw );
                                        Draw_Sprite( tmpbmp, lib.Get_Bitmap_At_Slot( fgslot ), 0, 0 );
                                        if Match( subbmp, tmpbmp ) then
                                            found_fg := not matched_fg;
                                            if not matched_fg then
                                                world.Set_Tile( 2, x, y, lib.Get_Id_At_Slot( fgslot ), notify => False );
                                            else
                                                -- found multiple foreground matches
                                                world.Set_Tile( 2, x, y, 0, notify => False );
                                            end if;
                                            found_bg := not matched_bg;
                                            if not matched_bg then
                                                world.Set_Tile( 1, x, y, lib.Get_Id_At_Slot( bgslot ), notify => False );
                                                matched_bg := True;
                                            else
                                                -- found multiple background matches
                                                world.Set_Tile( 1, x, y, 0, notify => False );
                                            end if;
                                        end if;
                                    end if;
                                end loop;
                                matched_fg := True;
                            end if;
                            tpos := Next( tpos );
                        end loop;
                    end if;

                    if found_fg or else found_bg then
                        matched := matched + 1;
                    end if;
                    Destroy_Bitmap( subbmp );
                end;

                if x mod 20 = 0 then
                    clk := Clock;
                    if clk - lastUpdate >= UPDATE_DELAY then
                        progress := 100 * (y * world.Get_Width_Tiles + x + 1) /
                                          (world.Get_Width_Tiles * world.Get_Height_Tiles);
                        if progress /= lastProgress then
                            secsLeft := Integer(Long_Float(100.0 / Float(progress)) *
                                                Long_Float(To_Duration(clk - start)) -
                                                Long_Float(To_Duration(clk - start)));
                            Update_Progress( progress, secsLeft );
                            lastProgress := progress;
                            lastUpdate := clk;
                        end if;
                    end if;
                end if;

            end loop;
        end loop;
        Update_Progress( 100, 0 );
        pragma Debug( Dbg( "Matched " & Image( matched ) & " of " &
                           Image( world.Get_Width_Tiles * world.Get_Height_Tiles ) &
                           " tiles: " & Image( 100 * matched / (world.Get_Width_Tiles * world.Get_Height_Tiles) ) & '%',
                           D_GAME, INFO ) );

        pragma Debug( matched = 0, Dbg( "Failed to match any tiles.", D_GAME, INFO ) );

        Destroy_Bitmap( tmpbmp );
        Destroy_Bitmap( bmp );
        Unload_Library( lib );
        return world;
    exception
        when e : others =>
            pragma Debug( Dbg( "Error importing world: " &
                               Exception_Name( e ) & " - " & Exception_Message( e ),
                               D_GAME, ERROR ) );
            Destroy_Bitmap( tmpbmp );
            Destroy_Bitmap( bmp );
            Unload_Library( lib );
            Delete( world );
            raise;
    end Import_World;

end Worlds.Importing;
