
--with Debugging;                         use Debugging;
with Events.World;                      use Events.World;
with Preferences;                       use Preferences;
with Worlds;

pragma Elaborate_All( Preferences );
pragma Elaborate_All( Worlds );

package body Worlds.Keen is

    READ_VERSION  : constant := 9;
    WRITE_VERSION : constant := 9;
    IDENTIFIER    : constant String := "keen7";

    ----------------------------------------------------------------------------

    function Create_World( width,
                           height  : Positive;
                           libName,
                           domain  : String ) return A_World is
        this : A_Keen_World := new Keen_World;
    begin
        this.Construct( width, height, libName, domain );
        return A_World(this);
    exception
        when others =>
            Delete( A_World(this) );
            raise;
    end Create_World;

    ----------------------------------------------------------------------------

    overriding
    procedure Construct( this : access Keen_World ) is
        map : A_Map := null;
    begin
        World_Object(this.all).Construct( "", map, "", "" );
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    procedure Construct( this    : access Keen_World;
                         width,
                         height  : Positive;
                         libName,
                         domain  : String ) is
        map    : A_Map := null;
        player : A_Entity := null;
    begin
        map := Create_Map( width, height );
        World_Object(this.all).Construct( "", map, libName, domain );
        player := Entities.Allocate( "player." & domain );
        if player = null then
            raise Constraint_Error with "Player allocator not registered for domain <" & domain & '>';
        end if;
        player.Set_World( this );
        player.Set_Location( Float(this.Get_Width / 2), Float(this.Get_Height / 2) );
        this.entities.Insert( player.Get_Id, player );
        this.player := player.Get_Id;
    exception
        when others =>
            Delete( player );
            raise;
    end Construct;

    ----------------------------------------------------------------------------

    not overriding
    function Get_Action( this : access Keen_World;
                         x, y : Integer ) return Action_Type is
        id     : Natural;
        tile   : A_Keen_Tile;
        action : Action_Type := NONE;
    begin
        if x >= 0 and then y >= 0 and then
           x < this.map.Get_Width and then y < this.map.Get_Height
        then
            for layer in 1..this.map.Get_Layers loop
                id := this.map.Get_Tile_Id( layer, x, y );
                if id /= 0 then
                    tile := A_Keen_Tile(this.lib.Get_Tile( id ));
                    if tile.Get_Action > action then
                        action := tile.Get_Action;
                    end if;
                end if;
            end loop;
        end if;
        return action;
    end Get_Action;

    ----------------------------------------------------------------------------

    overriding
    procedure Queue_Load_Events( this : access Keen_World ) is
    begin
        World_Object(this.all).Queue_Load_Events;
        Queue_World_Property_Changed( "introduction", To_String( this.introduction ) );
    end Queue_Load_Events;

    ----------------------------------------------------------------------------

    overriding
    procedure Set_Property( this : access Keen_World; name, value : String ) is
    begin
        if name = "introduction" then
            if value /= To_String( this.introduction ) then
                this.introduction := To_Unbounded_String( value );
                Queue_World_Property_Changed( name, value );
                Queue_World_Modified;
            end if;
        else
            World_Object(this.all).Set_Property( name, value );
        end if;
    end Set_Property;

    ----------------------------------------------------------------------------

    overriding
    function Object_Input( stream : access Root_Stream_Type'Class ) return Keen_World is
        this : aliased Keen_World;
    begin
        this.Construct;
        Keen_World'Read( stream, this );
        return this;
    end Object_Input;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Read( stream : access Root_Stream_Type'Class; obj : out Keen_World ) is
    begin
        World_Object'Read( stream, World_Object(obj) );
        obj.introduction := To_Unbounded_String( String'Input( stream ) );
    end Object_Read;

    ----------------------------------------------------------------------------

    overriding
    procedure Object_Write( stream : access Root_Stream_Type'Class; obj : Keen_World ) is
    begin
        World_Object'Write( stream, World_Object(obj) );
        String'Output( stream, To_String( obj.introduction ) );
    end Object_Write;

begin

    Register( Create_World'Access, IDENTIFIER, READ_VERSION, WRITE_VERSION );

    -- friction_x/y  : % velocity to maintain per tick
    -- gravity       : downward acceleration in pixels / second^2
    -- max_speed_x/y : maximum velocity in pixels / second

    Register_Domain( "Map" );
    Preferences.Set_Default( "physics", "Map.friction_x", 0.4 );
    Preferences.Set_Default( "physics", "Map.friction_y", 0.4 );
    Preferences.Set_Default( "physics", "Map.gravity", 0.0 );
    Preferences.Set_Default( "physics", "Map.max_speed_x", 1000.0 );
    Preferences.Set_Default( "physics", "Map.max_speed_y", 1000.0 );

    Register_Domain( "Platform" );
    Preferences.Set_Default( "physics", "Platform.friction_x", 0.6 );
    Preferences.Set_Default( "physics", "Platform.friction_y", 0.0 );
    Preferences.Set_Default( "physics", "Platform.gravity", 768.0 );
    Preferences.Set_Default( "physics", "Platform.max_speed_x", 140.0 );
    Preferences.Set_Default( "physics", "Platform.max_speed_y", 380.0 );

end Worlds.Keen;
