
with Ada.Characters.Handling;           use Ada.Characters.Handling;
with Ada.Numerics.Elementary_Functions; use Ada.Numerics.Elementary_Functions;
with Ada.Strings.Fixed;                 use Ada.Strings.Fixed;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with Ada.Text_IO;                       use Ada.Text_IO;
with Allegro.Bitmaps;                   use Allegro.Bitmaps;
with Allegro.Drawing;                   use Allegro.Drawing;
with Allegro.Image_Files;               use Allegro.Image_Files;
with Allegro.Sprites;                   use Allegro.Sprites;
with Allegro.Truecolor;                 use Allegro.Truecolor;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with LoadPNG;                           use LoadPNG;
with Resources;                         use Resources;
with Resources.Images;                  use Resources.Images;
with Support;                           use Support;
with Support.Paths;                     use Support.Paths;
with Themes.Palette;                    use Themes.Palette;

package body Tile_Operations is

    function Convert_BMP_To_24( filename : String ) return Boolean is
        cc  : constant Conv_Type := Get_Color_Conversion;
        ok  : Boolean := False;
        bmp : A_Bitmap;
    begin
        Set_Color_Conversion( 16#04FFFFFF# );
        bmp := Load_BMP( Normalize_Pathname( filename ) & ASCII.NUL, null );
        if bmp /= null then
#if WINDOWS then
            -- Save the file with a lower-case filename. This is acceptable on a
            -- non case sensitive file system like Windows. This is done because
            -- convert is primarily useful for converting old 8-bit resources
            -- that were created in the DOS days and have upper-case filenames.
            if Save_BMP( To_Lower( Normalize_Pathname( filename ) ), bmp, null ) then
#else
            if Save_BMP( Normalize_Pathname( filename ), bmp, null ) then
#end if;
                ok := True;
            end if;
            Destroy_Bitmap( bmp );
        end if;
        Set_Color_Conversion( cc );
        return ok;
    end Convert_BMP_To_24;

    ----------------------------------------------------------------------------

    procedure Cut( filename  : String;
                   width     : Positive;
                   spacing   : Natural := 0;
                   skipEmpty : Boolean := False;
                   prefix    : String := "";
                   format    : String := "";
                   indexPath : String := "";
                   startId   : Positive := 1;
                   textOut   : access procedure( str : String ) ) is
        tileId   : Positive := startId;
        idxFile  : File_Type;
        columns  : Integer;
        idDigits : Positive;

        ------------------------------------------------------------------------

        procedure Put_Line( str : String ) is
        begin
            if textOut /= null then
                textOut( "cut: " & str );
            end if;
        end Put_Line;

        ------------------------------------------------------------------------

        function Count_Digits( tileCount : Integer ) return Positive is
        begin
            if tileCount > 0 then
                return Integer(Float'Floor( Log( Float(tileCount), 10.0 ) )) + 1;
            end if;
            return 1;
        end Count_Digits;

        ------------------------------------------------------------------------

        function Is_Empty_Tile( bmp : not null A_Bitmap ) return Boolean is
            c : Color_Type;
        begin
            for y in 0..Get_Height( bmp )-1 loop
                for x in 0..Get_Width( bmp )-1 loop
                    c := Getpixel( bmp, x, y );
                    if c /= base_colors(Transparent) then
                        return False;
                    end if;
                end loop;
            end loop;
            return True;
        end Is_Empty_Tile;

        ------------------------------------------------------------------------

        procedure Save_Tile( outName : String; bmp : A_Bitmap ) is
            ok  : Boolean;
            pragma Warnings( Off, ok );
            ext : Unbounded_String;
        begin
            if format = "bmp" or else
               (format'Length = 0 and then To_Lower( Get_Extension( filename ) ) = "bmp")
            then
                ext := To_Unbounded_String( "bmp" );
                ok := Save_BMP( outName & ".bmp", bmp, null );
            elsif format = "png" or else
                  (format'Length = 0 and then To_Lower( Get_Extension( filename ) ) = "png")
            then
                ext := To_Unbounded_String( "png" );
                ok := (Save_PNG( outName & ".png", bmp, null ) = 0);
            end if;
            if Is_Open( idxFile ) then
                Put_Line( idxFile, String'((idDigits - Count_Digits( tileId )) * ' ') &
                                   Image( tileId ) & ' ' &
                                   Get_Filename( outName & '.' & ext ) );
            end if;
        end Save_Tile;

        ------------------------------------------------------------------------

        function Tile_Number( row, col : Integer ) return String is
            result : Unbounded_String;
        begin
            result := To_Unbounded_String( Image( row ) );
            if row < 10 then
                result := "0" & result;
            end if;
            if col < 10 then
                result := result & "0";
            end if;
            result := result & Image( col );
            return To_String( result );
        end Tile_Number;

        ------------------------------------------------------------------------

        res     : A_Resource_File;
        input   : A_Bitmap;
        output  : A_Bitmap;
    begin
        if not Is_Format_Supported( Get_Extension( filename ) ) then
            Put_Line( "Unsupported image format: " & Get_Extension( filename ) );
            return;
        end if;

        res := Load_Resource( Normalize_Pathname( filename ), "" );
        if res = null then
            Put_Line( "File not found: " & filename );
            return;
        end if;

        -- Load the raw file into memory
        input := Load_Image( res );
        Unload_Resource( res );
        if input = null then
            Put_Line( "Error loading image from " & filename );
            return;
        end if;

        begin
            -- Create the tile index file for writing. This will overwrite an
            -- existing file.
            if indexPath'Length > 0 then
                Create( idxFile, Out_File, Normalize_Pathname( indexPath ) );
            end if;
        exception
            when others =>
                Put_Line( "Error creating " & Normalize_Pathname( indexPath ) );
        end;

        output := Create_Bitmap( width, width );
        columns := Div_Ceil( Get_Width( input ), width + spacing );
        idDigits := Count_Digits( (startId - 1) +
                                  (columns *  Div_Ceil( Get_Height( input ), width + spacing )) );

        declare
            row  : Integer := 1;
            col  : Integer := 1;
            x, y : Integer;
        begin
            y := 0;
            while y < Get_Height( input ) loop
                x := 0;
                while x < Get_Width( input ) loop

                    -- Blit the cut out to the single tile bitmap
                    Blit( input, output, x, y, 0, 0, width, width );

                    -- Save the tile to disk
                    if not skipEmpty or else not Is_Empty_Tile( output ) then
                        Save_Tile( Normalize_Pathname( prefix & Tile_Number( row, col ) ), output );
                    elsif Is_Open( idxFile ) then
                        Put_Line( idxFile, String'((idDigits - 1) * ' ') & "0" );
                    end if;

                    -- Draw a row separator comment line
                    if ((tileId - startId) mod columns) = (columns - 1) then
                        Put_Line( idxFile, (80 * '-') );
                    end if;

                    tileId := tileId + 1;
                    col := col + 1;
                    x := x + width + spacing;
                end loop;
                col := 1;
                row := row + 1;
                y := y + width + spacing;
            end loop;
        end;

        Destroy_Bitmap( input );
        Destroy_Bitmap( output );
        if Is_Open( idxFile ) then
            Close( idxFile );
        end if;
    end Cut;

end Tile_Operations;
